"""
This module defines a device-independent MultiChannel Analyzer (MCA) class,
and a number of support classes.

Author:
   Mark Rivers
   
Created:
   Sept. 16, 2002.  Based on my earlier IDL code.

Modifications:
   Sept. 24, 2002 MLR
      - Fixed bug in saving ROIs in Meds
      - Fixed bug reading environment variables
   Sept. 25, 2002
      - Changed McaPeak.get_calibration to McaPeak.update, make more fields
        consistent.
      - Added McaPeak.ignore field to fix problem with other fields getting
        clobbered in fitPeaks.
      - Fixed serious bug in .d_to_channel()
   SAH - June 2019.
     - removed background and peak fitting functions because not required in integrated code.
     - removed import of CARSmath Xrf and fitPeaks
"""

import numpy as Numeric
#import string
import copy
import math
import sys
import time
#import CARSMath
#import Xrf
#import fitPeaks

########################################################################
class McaBackground:
   """
   Defines the parameters for fitting backgrounds in Mca objects.
   Fields and default values:
      .exponent     = 2
      .top_width    = 0.
      .bottom_width = 4.
      .tangent      = 0
      .compress     = 4
      
   See the documentation on fit_background() for information on the meaning
   of these fields.
   """
   def __init__(self):
      self.exponent     = 2
      self.top_width    = 0.
      self.bottom_width = 4.
      self.tangent      = 0
      self.compress     = 4

########################################################################
class McaFit:
   """
   Class used for the global input and output parameters for fit_peaks.
   """
   def __init__(self, mca=None):
      if (mca == None): mca=Mca()
      calibration = mca.get_calibration()
      self.npeaks =       0      # Number of peaks to fit
      self.first_chan =   0      # First channel to fit
      self.nchans =       len(mca.data)   # Number of channels 
      self.last_chan =    self.nchans-1   # Last channel to fit
      self.nparams =      0      # Number of fit parameters
      self.initial_energy_offset = calibration.offset
      self.initial_energy_slope =  calibration.slope
      self.energy_offset =  0.   # Energy calibration offset
      self.energy_slope =   0.   # Energy calibration slope
      self.energy_flag =    1    # Energy flag
                            #   0 = Fix energy calibration coefficients
                            #   1 = Optimize energy calibration coefficients
      self.initial_fwhm_offset = .15 # FWHM offset = 150 eV
      self.initial_fwhm_slope =  0.
      self.fwhm_offset =    0.   # FWHM offset
      self.fwhm_slope =     0.   # FWHM slope
      self.fwhm_flag =      1    # Fwhm flag
                            #   0 = Fix FWHM coefficients
                            #   1 = Optimize FWHM coefficients
      self.chi_exp =      0.     # Exponent of chi
      self.max_eval =     0      # Maximum # function evaluations
      self.n_eval =       0      # Actual number of function evalutions
      self.max_iter =     20     # Maximum number of iterations
      self.n_iter =       0      # Actual number of iterations
      self.tolerance =    1.e-4  # Convergence tolerance
      self.chisqr =       0.     # Chi-squared on output
      self.status =       0      # Output status code
      self.err_string =   ''     # Output error string
      self.debug =        0      # 0 (no debug) or 1
      self.results_file = 'fit_results.txt'  # Name of file for results
      self.spreadsheet_file = 'fit_spreadsheet.txt'  # Name of spreadsheet
      self.background =   McaBackground()  # Background object
      self.peaks =        []     # List of McaPeak() objects

   def update(self, mca):
      """
      Updates the McaFit object to be consistent with the Mca object.
      The calibration offset and slope are copied to the initial values,
      and the number of channels is set.

      Inputs:
         mca:
            An Mca object to copy the calibration from
      """
      cal = mca.get_calibration()
      self.initial_energy_offset = cal.offset
      self.initial_energy_slope = cal.slope
      data = mca.get_data()
      self.nchans = len(data)
      self.last_chan = self.nchans-1

########################################################################
class McaPeak:
   """
   Class for definiing the input and output parameters for each peak in
   fit_peaks().
   Input fields set bfore calling fit_peaks(), defaults and descriptions
      .label =       ""      # Peak label
      .self.energy_flag = 0  # Flag for fitting energy
                             #   0 = Fix energy 
                             #   1 = Optimize energy
      .fwhm_flag =   0       # Flag for fitting FWHM
                             #   0 = Fix FWHM to global curve
                             #   1 = Optimize FWHM
                             #   2 = Fix FWHM to input value
      .ampl_factor = 0.      # Fixed amplitude ratio to previous peak
                             #   0.  = Optimize amplitude of this peak
                             #   >0. = Fix amplitude to this value relative
                             #         to amplitude of previous free peak
                             #  -1.0 = Fix amplitude at 0.0
      .initial_energy = 0.   # Peak energy
      .initial_fwhm =   0.   # Peak FWHM
      .initial_ampl =   0.   # Peak amplitude
      
   Output fields returned by fit_peaks(), defaults and descriptions
      .energy =         0.   # Peak energy
      .fwhm =           0.   # Peak FWHM
      .ampl =           0.   # Peak amplitude
      .area =           0.   # Area of peak
      .bgd =            0.   # Background under peak
   """
   def __init__(self):
      self.label =       ""  # Peak label
      self.energy_flag = 0   # Flag for fitting energy
                             #   0 = Fix energy 
                             #   1 = Optimize energy
      self.fwhm_flag =   0   # Flag for fitting FWHM
                             #   0 = Fix FWHM to global curve
                             #   1 = Optimize FWHM
                             #   2 = Fix FWHM to input value
      self.ampl_factor = 0.  # Fixed amplitude ratio to previous peak
                             #   0.  = Optimize amplitude of this peak
                             #   >0. = Fix amplitude to this value relative
                             #         to amplitude of previous free peak
                             #  -1.0 = Fix amplitude at 0.0
      self.initial_energy = 0.  # Peak energy
      self.energy =         0.  # Peak energy
      self.initial_fwhm =   0.  # Peak FWHM
      self.fwhm =           0.  # Peak FWHM
      self.initial_ampl =   0.  # Peak amplitude
      self.ampl =           0.  # Peak amplitude
      self.area =           0.  # Area of peak
      self.bgd =            0.  # Background under peak
      self.ignore =         0   # Don't fit peak

########################################################################
class McaROI:
   """
   Class that defines a Region-Of-Interest (ROI)
   Fields
      .left      # Left channel or energy
      .right     # Right channel or energy
      .centroid  # Centroid channel or energy
      .fwhm      # Width
      .bgd_width # Number of channels to use for background subtraction
      .use       # Flag: should the ROI should be used for energy calibration
      .preset    # Is this ROI controlling preset acquisition
      .label     # Name of the ROI
      .d_spacing # Lattice spacing if a diffraction peak
      .energy    # Energy of the centroid for energy calibration
   """
   def __init__(self, left=0., right=0., centroid=0., fwhm=0., bgd_width=0,
                      use=1, preset=0, label='', d_spacing=0., energy=0.):
      """
      Keywords:
         There is a keyword with the same name as each attribute that can be
         used to initialize the ROI when it is created.
      """
      self.left = left
      self.right = right
      self.centroid = centroid
      self.fwhm = fwhm
      self.bgd_width = bgd_width
      self.use = use
      self.preset = preset
      self.label = label
      self.d_spacing = d_spacing
      self.energy = energy
   def __cmp__(self, other):
      """
      Comparison operator.  The .left field is used to define ROI ordering
      """ 
      return (self.left - other.left)

########################################################################
class McaCalibration:
   """
   Class defining an Mca calibration.  The calibration equation is
      energy = .offset + .slope*channel + .quad*channel**2
   where the first channel is channel 0, and thus the energy of the first
   channel is .offset.
   
   Fields:
      .offset    # Offset
      .slope     # Slope
      .quad      # Quadratic
      .units     # Calibration units, a string
      .two_theta # 2-theta of this Mca for energy-dispersive diffraction
   """
   def __init__(self, offset=0., slope=1.0, quad=0., units='keV', 
                      two_theta=10.):
      """
      There is a keyword with the same name as each field, so the object can
      be initialized when it is created.
      """
      self.offset = offset
      self.slope = slope
      self.quad = quad
      self.units = units
      self.two_theta = two_theta

########################################################################
class McaElapsed:
   """
   The elapsed time and counts for an Mca.
   
   Fields:
      .start_time   # Start time and date, a string
      .live_time    # Elapsed live time in seconds
      .real_time    # Elapsed real time in seconds
      .read_time    # Time that the Mca was last read in seconds
      .total_counts # Total counts between the preset start and stop channels
   """
   def __init__(self, start_time='', live_time=0., real_time=0., 
                      read_time=0., total_counts=0.):
      self.start_time = start_time
      self.live_time = live_time
      self.real_time = real_time
      self.read_time = read_time
      self.total_counts = total_counts

########################################################################
class McaPresets:
   """
   The preset time and counts for an Mca.
   
   Fields:
      .live_time       # Preset live time in seconds
      .real_time       # Preset real time in seconds
      .read_time       # Time that the Mca was last read in seconds
      .total_counts    # Preset total counts between the preset
                       #    start and stop channels
      .start_channel   # Start channel for preset counts
      .end_channel     # End channel for preset counts
      .dwell           # Dwell time per channel for MCS devices
      .channel_advance # Channel advance source for MCS hardware:
                       #    0=internal, 1=external
      .prescale        # Prescaling setting for MCS hardware
   """
   def __init__(self):
      self.live_time = 0.
      self.real_time = 0.
      self.total_counts = 0.
      self.start_channel = 0
      self.end_channel = 0
      self.dwell = 0.
      self.channel_advance = 0
      self.prescale = 0

########################################################################
class McaEnvironment:
   """
   The "environment" or related parameters for an Mca.  These might include
   things like motor positions, temperature, anything that describes the
   experiment.

   An Mca object has an associated list of McaEnvironment objects, since there
   are typically many such parameters required to describe an experiment.

   Fields:
      .name         # A string name of this parameter, e.g. "13IDD:m1"
      .value        # A string value of this parameter,  e.g. "14.223"
      .description  # A string description of this parameter, e.g. "X stage"
   """
   def __init__(self, name='', value='', description=''):
      self.name = name
      self.value = value
      self.description = description

########################################################################
class Mca(object):
   """ Device-independent MultiChannel Analyzer (MCA) class """
   def __init__(self, file=None, **filekw):
      """
      Creates new Mca object.  The data are initially all zeros, and the number
      of channels is 2048.
      
      Keywords:
         file:
            Name of a file to read into the new Mca object with read_file()
            
      Example:
         m = Mca('my_spectrum.dat')
      """
      self.name = ''
      self.n_detectors = 1
      nchans = 2048
      self.data = Numeric.zeros(nchans)
      self.rois = []
      self.calibration = McaCalibration()
      self.elapsed = McaElapsed()
      self.presets = McaPresets()
      self.environment = []
      if (file != None):
         self.read_file(file, **filekw)

   ########################################################################
   def __copy__(self):
      """
      Makes a "shallow" copy of an Mca instance, using copy.copy() on all of
      the attributes of the Mca instance.  The .rois and .environment attributes
      will still point to the same values, because they are lists.
      """
      new = Mca()
      new.data = copy.copy(self.data)
      new.rois = copy.copy(self.rois)
      new.elapsed = copy.copy(self.elapsed)
      new.calibration = copy.copy(self.calibration)
      new.presets = copy.copy(self.presets)
      new.environment = copy.copy(self.environment)
      return(new)


   ########################################################################
   def __deepcopy__(self, visit):
      """
      Makes a "deep" copy of an Mca instance, using copy.deepcopy() on all of
      the attributes of the Mca instance. All of the attribute will point to
      new objects.
      """
      new = Mca()
      new.data = copy.copy(self.data)
      new.rois = copy.deepcopy(self.rois, visit)
      new.elapsed = copy.copy(self.elapsed)
      new.calibration = copy.copy(self.calibration)
      new.presets = copy.copy(self.presets)
      new.environment = copy.deepcopy(self.environment, visit)
      return(new)

   ########################################################################
   def get_calibration(self):
      """ Returns the Mca calibration, as an McaCalibration object """
      return self.calibration

   ########################################################################
   def set_calibration(self, calibration):
      """
      Sets the Mca calibration.
      
      Inputs:
         calibration:
            An McaCalibration object
      """
      self.calibration = calibration

   ########################################################################
   def get_presets(self):
      """ Returns the Mca presets, as an McaCPresets object """
      return self.presets

   ########################################################################
   def set_presets(self, presets):
      """
      Sets the Mca presets.
      
      Inputs:
         presets:
            An McaPresets object
      """
      self.presets = presets

   ########################################################################
   def get_elapsed(self):
      """ Returns the Mca elapsed parameters, as an McaElapsed object """
      return self.elapsed

   ########################################################################
   def set_elapsed(self, elapsed):
      """
      Sets the Mca elapsed parameters.
      
      Inputs:
         elapsed:
            An McaElapsed object
      """
      self.elapsed = elapsed

   ########################################################################
   def get_name(self):
      """ Returns the Mca name as a string """
      return self.name

   ########################################################################
   def set_name(self, name):
      """
      Sets the Mca name.
      
      Inputs:
         name:
            A string
      """
      self.name = name

   ########################################################################
   def get_rois(self, energy=0):
      """ Returns the Mca ROIS, as a list of McaROI objects """
      rois = copy.copy(self.rois)
      if (energy != 0):
         for roi in rois:
            roi.left = self.channel_to_energy(roi.left)
            roi.right = self.channel_to_energy(roi.right)
      return rois

   ########################################################################
   def set_rois(self, rois, energy=0):
      """
      Sets the region-of-interest parameters for the MCA.
      The rois information is contained in an object of class McaRoi.
      This routine is not needed if the information in the McaRoi instance
      is already in channel units.  It is needed if the information in the
      .left and .right fields is in terms of energy.

      Inputs:
         rois:
            A list of objects of type McaROI
            
      Keywords:
         energy:
            Set this flag to indicate that the .left and .right fields
            of rois are in units of energy rather than channel number.
            
      Example:
        mca = Mca('mca.001')
        r1 = McaROI()
        r1.left = 5.4
        r1.right = 5.6
        r2 = McaROI()
        r2.left = 6.1
        r2.right = 6.2
        mca.set_rois([r1,r2], energy=1)
      """
      self.rois = []
      for roi in rois:
         r = roi
         if (energy == 1):
            r.left =  self.energy_to_channel(r.left, clip=1)
            r.right = self.energy_to_channel(r.right, clip=1)
         self.rois.append(r)

   ########################################################################
   def get_roi_counts(self, background_width=1):
      """
      Returns a tuple (total, net) containing the total and net counts of
      each region-of-interest in the MCA.

      Kyetwords:
         background_width:
            Set this keyword to set the width of the background region on either
            side of the peaks when computing net counts.  The default is 1.
            
      Outputs:
          total:  The total counts in each ROI.
          net:    The net counts in each ROI.

          The dimension of each list is NROIS, where NROIS
          is the number of currently defined ROIs for this MCA.  It returns
          and empty list for both if NROIS is zero.
          
      Example:
         mca = Mca('mca.001')
         total, net = mca.get_roi_counts(background_width=3)
         print 'Net counts = ', net
      """
      total = []
      net = []
      nchans = len(self.data)
      for roi in self.rois:
         left = roi.left
         ll = max((left-background_width+1), 0)
         if (background_width > 0):
             bgd_left = sum(self.data[ll:(left+1)]) / (left-ll+1)
         else: bgd_left = 0.
         right = roi.right
         rr = min((right+background_width-1), nchans-1)
         if (background_width > 0):
             bgd_right = sum(self.data[right:rr+1]) / (rr-right+1)
         else: bgd_right = 0.
         total_counts = self.data[left:right+1]
         total.append(sum(total_counts))
         n_sel        = right - left + 1
         bgd_counts   = bgd_left + Numeric.arange(n_sel,Float)/(n_sel-1) * \
                                  (bgd_right - bgd_left)
         net_counts   = total_counts - bgd_counts
         net.append(sum(net_counts))
      return (total, net)

   ########################################################################
   def get_environment(self):
      """
      Returns a list of McaEnvironment objects that contain the environment
      parameters of the Mca.
      """
      return self.environment

   ########################################################################
   def set_environment(self, environment):
      """
      Copies a list of McaEnvironment objects to the Mca object.

      Inputs:
         environment:
            A list of McaEnvironment objects.
      """
      self.environment = environment

   ########################################################################
   def get_data(self):
      """ Returns the data (counts) from the Mca """
      return self.data

   ########################################################################
   def set_data(self, data):
      """
      Copies an array of data (counts) to the Mca.

      Inputs:
         data:
            A Numeric array of data (counts).
      """
      self.data = data

   ########################################################################
   def get_energy(self):
      """
      Returns a list containing the energy of each channel in the MCA spectrum.

      Procedure:
         Simply returns mca.channel_to_energy() for each channel
         
      Example:
          from Mca import *
          mca = Mca('mca.001')
          energy = mca.get_energy()
      """
      channels = Numeric.arange(len(self.data))
      return self.channel_to_energy(channels)

   ########################################################################
   def initial_calibration(self, energy):
      """
      Performs an initial coarse energy calibration of the Mca, setting only
      the slope, and setting the offset parameter to 0.

      Inputs:
         energy: The energy of the biggest peak in the MCA spectrum.
         
      Procedure:
         This routine does the following:
            1) Sets the offset coefficient to 0.0
            2) Sets the quadratic coefficient to 0.0
            3) Determines which channel contains the most counts, PEAK_CHAN
            4) Sets the slope equal to the input energy divided by PEAK_CHAN
            
      Example:
         from Mca import *
         mca = Mca('mca.001')
         mca.initial_calibration(20.1)
      """
      peak_chan = Numeric.argmax(self.data)
      peak_chan = max(peak_chan,1)
      self.calibration.offset = 0.
      self.calibration.slope = float(energy)/peak_chan
      self.calibration.quad = 0.

   ########################################################################
   def add_roi(self, roi, energy=0):
      """
      This procedure adds a new region-of-interest to the MCA.

      Inputs:
         roi: An object of type mcaROI.
         
      Kyetwords:
         energy:
            Set this flag to 1 to indicate that the .left and .right 
            fields of roi are in units of energy rather than channel 
            number.
            
      Example:
         mca = Mca('mca.001')
         roi = McaROI()
         roi.left = 500
         roi.right = 600
         roi.label = 'Fe Ka'
         mca,add_roi(roi)
      """
      r = copy.copy(roi)
      if (energy == 1):
         r.left = self.energy_to_channel(r.left, clip=1)
         r.right = self.energy_to_channel(r.right, clip=1)
      self.rois.append(r)

      # Sort ROIs.  This sorts by left channel.
      self.rois.sort()

   ########################################################################
   def find_roi(self, left, right, energy=0):
      """
      This procedure finds the index number of the ROI with a specified
      left and right channel number.

      Inputs:
         left:
            Left channel number (or energy) of this ROI
            
         right:
            Right channel number (or energy) of this ROI
         
      Keywords:
         energy:
            Set this flag to 1 to indicate that Left and Right are in units
            of energy rather than channel number.
            
      Output:
         Returns the index of the specified ROI, -1 if the ROI was not found.
         
      Example:
         mca = Mca('mca.001')
         index = mca.find_roi(100, 200)
      """
      l = left
      r = right
      if (energy == 1):
         l = self.energy_to_channel(l, clip=1)
         r = self.energy_to_channel(r, clip=1)
      index = 0
      for roi in self.rois:
         if (l == roi.left) and (r == roi.right): return index
         index = index + 1
      return -1

   ########################################################################
   def delete_roi(self, index):
      """
      This procedure deletes the specified region-of-interest from the MCA.

      Inputs:
         index:  The index of the ROI to be deleted, range 0 to len(mca.rois)
         
      Example:
        mca = Mca('mca.001')
        mca.delete_roi(2)
      """
      del self.rois[index]

   ########################################################################
   def channel_to_energy(self, channels):
      """
      Converts channels to energy using the current calibration values for the
      Mca.  This routine can convert a single channel number or an array of
      channel numbers.  Users are strongly encouraged to use this function
      rather than implement the conversion calculation themselves, since it
      will be updated if additional calibration parameters (cubic, etc.) are
      added.

      Inputs:
         channels:
            The channel numbers to be converted to energy.  This can be
            a single number or a sequence of channel numbers.
            
      Outputs:
         This function returns the equivalent energy for the input channels.
         
      Example:
         mca = Mca('mca.001')
         channels = [100, 200, 300]
         energy = mca.channel_to_energy(channels) # Get the energy of these
      """

      c = Numeric.asarray(channels)
      return self.calibration.offset + \
             self.calibration.slope * c + \
             self.calibration.quad * Numeric.power(c, 2)

   ########################################################################
   def channel_to_d(self, channels):
      """
      Converts channels to "d-spacing" using the current calibration values for
      the Mca.  This routine can convert a single channel number or an array of
      channel numbers.  Users are strongly encouraged to use this function
      rather than implement the conversion calculation themselves, since it
      will be updated if additional calibration parameters are added.  This
      routine is useful for energy dispersive diffraction experiments.  It uses
      both the energy calibration parameters and the "two-theta" calibration
      parameter.

      Inputs:
         channels:
            The channel numbers to be converted to "d-spacing".
            This can be a single number or a list of channel numbers.
            
      Outputs:
         This function returns the equivalent "d-spacing" for the input channels.
         The output units are in Angstroms.
         
      Restrictions:
         This function assumes that the units of the energy calibration are keV
         and that the units of "two-theta" are degrees.
         
      Example:
         mca = Mca('mca.001')
         channels = [100,200,300]
         d = mca.channel_to_d(channels)       # Get the "d-spacing" of these
      """
      e = self.channel_to_energy(channels)
      return 12.398 / (2. * e * math.sin(self.calibration.two_theta/2.*math.pi/180.))
  
    #FIXED : added divide by 2 to conversion

   ########################################################################
   def energy_to_channel(self, energy, clip=0):
      """
      Converts energy to channels using the current calibration values for the
      Mca.  This routine can convert a single energy or an array of energy
      values.  Users are strongly encouraged to use this function rather than
      implement the conversion calculation themselves, since it will be updated
      if additional calibration parameters are added.

      Inputs:
         energy:
            The energy values to be converted to channels. This can be a
            single number or a sequence energy values.
            
      Keywords:
         clip:
            Set this flag to 1 to clip the returned values to be between
            0 and nchans-1.  The default is not to clip.
            
      Outputs:
         This function returns the closest equivalent channel for the input
         energy.  Note that it does not generate an error if the channel number
         is outside the range 0 to (nchans-1), which will happen if the energy
         is outside the range for the calibration values of the Mca.
         
      Example:
         mca = Mca('mca.001')
         channel = mca.energy_to_channel(5.985)
      """
      if (self.calibration.quad == 0.0):
         channel = ((energy-self.calibration.offset) /
                    self.calibration.slope)
      else:
         # Use the quadratic formula, use some shorthand
         a = self.calibration.quad
         b = self.calibration.slope
         c = self.calibration.offset - energy
         # There are 2 roots.  I think we always want the "+" root?
         channel = (-b + Numeric.sqrt(b**2 - 4.*a*c))/(2.*a)
      channel = Numeric.around(channel)
      if (clip != 0): 
         nchans = len(self.data)
         channel = Numeric.clip(channel, 0, nchans-1)
      if (type(channel) == Numeric.ArrayType): 
         return channel.astype(Numeric.Int)
      else:
         return int(channel)

   ########################################################################
   def d_to_channel(self, d, clip=0):
      """
      Converts "d-spacing" to channels using the current calibration values
      for the Mca.  This routine can convert a single "d-spacing" or an array
      of "d-spacings".  Users are strongly encouraged to use this function
      rather than implement the conversion calculation themselves, since it
      will be updated if additional calibration parameters are added.
      This routine is useful for energy dispersive diffraction experiments.
      It uses both the energy calibration parameters and the "two-theta"
      calibration parameter.

      Inputs:
         d:
            The "d-spacing" values to be converted to channels.
            This can be a single number or an array of values.
            
      Keywords:
         clip:
            Set this flag to 1 to clip the returned values to be between
            0 and nchans-1.  The default is not to clip.
            
      Outputs:
         This function returns the closest equivalent channel for the input
         "d-spacing". Note that it does not generate an error if the channel
         number is outside the range 0 to (nchans-1), which will happen if the
         "d-spacing" is outside the range for the calibration values of the Mca.
         
      Example:
         mca = Mca('mca.001')
         channel = mca.d_to_chan(1.598)
      """
      e = 12.398 / (2. * d * math.sin(self.calibration.two_theta*math.pi/180./2.))
      return self.energy_to_channel(e, clip=clip)

   ########################################################################
   def write_file(self, file, netcdf=0):
      """
      Writes Mca or Med objects to a disk file.
      
      It calls Mca.write_netcdf_file if the netcdf keyword flg is set,

      Note that users who want to read such files with Python are strongly
      encouraged to use Mca.read_file()

      Inputs:
         file:
            The name of the disk file to write.
            
      Keywords:
         netcdf:
            Set this flag to write the file in netCDF format, otherwise
            the file is written in ASCII format.  See the documentation
            for Mca.write_ascii_file and Mca.write_netcdf_file for 
            information on the formats.
 
      Example:
         mca = Mca()
         mca.write_file('mca.001')
      """
      # Call the get_xxx() methods to make sure things are up to date
      data = self.get_data()
      calibration = self.get_calibration()
      elapsed = self.get_elapsed()
      presets = self.get_presets()
      rois = self.get_rois()
      environment = self.get_environment()

      if (netcdf != 0):
        write_netcdf_file(file, data, calibration, elapsed, presets, rois,
                          environment)
      else:
        write_ascii_file(file, data, calibration, elapsed, presets, rois,
                         environment)



   ########################################################################
   def read_file(self, file, netcdf=0, detector=0):
      """
      Reads a disk file into an MCA object.  If the netcdf=1 flag is set it
      reads a netcdf file, else it assumes the file is ASCII.
      If the data file has multiple detectors then the detector keyword can be
      used to specify which detector data to return.

      Inputs:
         file:
            The name of the disk file to read.
            
      Keywords:
         netcdf:
            Set this flag to read files written in netCDF format, otherwise
            the routine assumes that the file is in ASCII format.
            See the documentation for Mca.write_ascii_file and
            Mca.write_netcdf_file for information on the formats.

         detector:
            Specifies which detector to read if the file has multiple detectors.
            
      Example:
         mca = Mca()
         mca.read_file('mca.001')
      """
      if (netcdf != 0):
         r = read_netcdf_file(file)
      else:
         r = read_ascii_file(file)
      self.name = file
      self.calibration = r['calibration'][detector]
      self.data = r['data'][detector]
      self.elapsed = r['elapsed'][detector]
      self.rois = r['rois'][detector]
      self.environment = r['environment']



   ########################################################################
#   def fit_background(self, bottom_width=4., top_width=0., exponent=2, 
#                      tangent=0, compress=4):
#      """
#      This function fits a background to an MCA spectrum. The background is
#      fitted using an enhanced version of the algorithm published by
#      Kajfosz, J. and Kwiatek, W .M. (1987)  "Non-polynomial approximation of
#      background in x-ray spectra." Nucl. Instrum. Methods B22, 78-81.
#      
#      Keywords:
#         top_width:
#            Specifies the width of the polynomials which are concave upward.
#            The top_width is the full width in energy units at which the
#            magnitude of the polynomial is 100 counts. The default is 0, which
#            means that concave upward polynomials are not used.
#
#         bottom_width:
#            Specifies the width of the polynomials which are concave downward.
#            The bottom_width is the full width in energy units at which the
#            magnitude of the polynomial is 100 counts. The default is 4.
#
#         exponent:
#            Specifies the power of polynomial which is used. The power must be
#            an integer. The default is 2, i.e. parabolas. Higher exponents,
#            for example EXPONENT=4, results in polynomials with flatter tops
#            and steeper sides, which can better fit spectra with steeply
#            sloping backgrounds.
#
#         tangent:
#            Specifies that the polynomials are to be tangent to the slope of the
#            spectrum. The default is vertical polynomials. This option works
#            best on steeply sloping spectra. It has trouble in spectra with
#            big peaks because the polynomials are very tilted up inside the
#            peaks.
#
#         compress:
#            Compression factor to apply before fitting the background.
#            Default=4, which means, for example, that a 2048 channel spectrum
#            will be rebinned to 512 channels before fitting.
#            The compression is done on a temporary copy of the input spectrum,
#            so the input spectrum itself is unchanged.
#            The algorithm works best if the spectrum is compressed before it
#            is fitted. There are two reasons for this. First, the background
#            is constrained to never be larger than the data itself. If the
#            spectrum has negative noise spikes they will cause the fit to be
#            too low. Compression will smooth out such noise spikes.
#            Second, the algorithm requires about 3*N^2 operations, so the time
#            required grows rapidly with the size of the input spectrum. On a
#            200 MHz Pentium it takes about 3 seconds to fit a 2048 channel
#            spectrum with COMPRESS=1 (no compression), but only 0.2 seconds
#            with COMPRESS=4 (the default).
#            
#     Procedure:
#         1) At each channel "i" an n'th degree polynomial which is concave up
#         is fitted. Its equation is
#
#                                       n
#                          (e(i) - e(j))
#         f(j,i) = y(i) + --------------
#                                    n
#                           top_width
#
#         where f(j,i) is the fitted counts in channel j for the polynomial
#         centered in channel i. y(i) is the input counts in channel "i", e(i) is
#         the energy of channel i, e(j) is the energy of channel j, and
#         "top_width" and "n" are user-specified parameters. The background count
#         in channel "j", b(j) is defined as
#
#         b(j) = min ((f(j,i), y(j))
#                 i
#
#         b(j) is thus the smallest fitted polynomial in channel j, or the raw
#         data, whichever is smaller.
#
#         2) After the concave up polynomials have been fitted, a series of
#         concave down polynomials are constructed. At each channel "i" an n'th
#         degree polynomial which is concave up is fitted. The polynomial is slid
#         up from below until it "just touches" some channel of the spectrum. Call
#         this channel "i". The maximum height of the polynomial is thus
#
#                                                  n
#                                     (e(i) - e(j))
#         height(j) = max ( b(j) +  --------------  )
#                      i                          n
#                                     bottom_width
#
#         where bottom_width is a user_specified parameter.
#
#         3) Once the value of height(i) is known the polynomial is fitted. The
#         background counts in each channel are then determined from:
#
#                                                  n
#                                     (e(i) - e(j))
#         bgd(j) = max ( height(i) + --------------
#                   i                             n
#                                     bottom_width
#
#         bgd(j) is thus the maximum counts for any of the concave down
#         polynomials passing though channel j.
#
#         Before the concave-down polynomials are fitted the spectrum at each
#         channel it is possible to subtract out a straight line which is
#         tangent to the spectrum at that channel. Use the /TANGENT qualifier to
#         do this. This is equivalent to fitting a "tilted" polynomial whose
#         apex is tangent to the spectrum at that channel. By fitting
#         polynomials which are tangent rather than vertical the background fit
#         is much improved on spectra with steep slopes.
#
#      Outputs:
#         This function returns an MCA object which is identical to the calling
#         object, except that the data have been replaced by the background fit.
#
#      Example:
#        mca = Mca()
#        mca.read_file('mca.001')
#        bgd = mca.fit_background(mca, bottom=6, exponent=4)
#      """
#      REFERENCE_AMPL=100.
#      TINY = 1.E-20
#      HUGE = 1.E20
#      MAX_TANGENT=2
#
#      bgd = copy.copy(self)
#      nchans = len(bgd.data)
#      calibration = bgd.get_calibration()
#      scratch = copy.copy(bgd.get_data())
#      slope = calibration.slope
#
#      # Compress scratch spectrum
#      if (compress > 1):
#         scratch = CARSMath.compress_array(scratch, compress)
#         slope = slope * compress
#         nchans = nchans / compress
#
#      # Copy scratch spectrum to background spectrum
#      bckgnd = copy.copy(scratch)
#
#      # Find maximum counts in input spectrum. This information is used to
#      # limit the size of the function lookup table
#      max_counts = max(scratch)
#
#      #  Fit functions which come down from top
#      if (top_width > 0.):
#         #   First make a lookup table of this function
#         chan_width = top_width / (2. * slope)
#         denom = chan_width**exponent
#         indices = Numeric.arange(float(nchans*2+1)) - nchans
#         power_funct = indices**exponent * (REFERENCE_AMPL / denom)
#         power_funct = Numeric.compress((power_funct <= max_counts), power_funct)
#         max_index = len(power_funct)/2 - 1
#
#         for center_chan in range(nchans):
#            first_chan = max((center_chan - max_index), 0)
#            last_chan = min((center_chan + max_index), (nchans-1))
#            f = first_chan - center_chan + max_index
#            l = last_chan - center_chan + max_index
#            test = scratch[center_chan] + power_funct[f:l+1]
#            sub = bckgnd[first_chan:last_chan+1] 
#            bckgnd[first_chan:last_chan+1] = Numeric.maximum(sub, test)
#
#      # Copy this approximation of background to scratch
#      scratch = copy.copy(bckgnd)
#
#      # Find maximum counts in scratch spectrum. This information is used to
#      #   limit the size of the function lookup table
#      max_counts = max(scratch)
#
#      # Fit functions which come up from below
#      bckgnd = Numeric.arange(float(nchans)) - HUGE
#
#      # First make a lookup table of this function
#      chan_width = bottom_width / (2. * slope)
#      if (chan_width == 0.): denom = TINY
#      else: denom = chan_width**exponent
#      indices = Numeric.arange(float(nchans*2+1)) - nchans
#      power_funct = indices**exponent  * (REFERENCE_AMPL / denom)
#      power_funct = Numeric.compress((power_funct <= max_counts), power_funct)
#      max_index = len(power_funct)/2 - 1
#
#      for center_chan in range(nchans-1):
#         tangent_slope = 0.
#         if (tangent):
#            # Find slope of tangent to spectrum at this channel
#            first_chan = max((center_chan - MAX_TANGENT), 0)
#            last_chan = min((center_chan + MAX_TANGENT), (nchans-1))
#            denom = center_chan - Numeric.arange(float(last_chan - first_chan + 1))
#            tangent_slope = (scratch[center_chan] - 
#                             scratch[first_chan:last_chan+1]) / max(denom, 1)
#            tangent_slope = Numeric.sum(tangent_slope) / (last_chan - first_chan)
#
#         first_chan = max((center_chan - max_index), 0)
#         last_chan = min((center_chan + max_index), (nchans-1))
#         last_chan = max(last_chan, first_chan)
#         nc = last_chan - first_chan + 1
#         lin_offset = scratch[center_chan] + \
#                     (Numeric.arange(float(nc)) - nc/2) * tangent_slope
#
#         # Find the maximum height of a function centered on this channel
#         # such that it is never higher than the counts in any channel
#
#         f = first_chan - center_chan + max_index
#         l = last_chan - center_chan + max_index
#         test = scratch[first_chan:last_chan+1] - lin_offset + \
#                                                            power_funct[f:l+1]
#         height = min(test)
#
#         # We now have the function height. Set the background to the
#         # height of the maximum function amplitude at each channel
#
#         test = height + lin_offset - power_funct[f:l+1]
#         sub = bckgnd[first_chan:last_chan+1]
#         bckgnd[first_chan:last_chan+1] = Numeric.maximum(sub, test)
#
#      # Expand spectrum
#      if (compress > 1): bckgnd = CARSMath.expand_array(bckgnd, compress)
#      bgd.set_data(bckgnd.astype(Numeric.Int))
#      return bgd
#

   ########################################################################
#   def fit_peaks(self, peaks, fit=None, background=None,
#                 output='', spreadsheet=None, 
#                 append=1, **background_kw):
#      """
#      Fits the peaks in the MCA spectrum. It provides a convenient interface to
#      the fitPeaks() function.
#   
#      Inputs:
#         peaks:  A list of McaPeak objects.  See fitPeaks() and read_peaks() for
#         more information.
#
#      Keywords:
#         fit:
#            An object of type McaFit which can be used to control the
#            peak fitting.  If this keyword is omitted then the fit structure
#            is created with McaFit()
#            
#         background:
#            An Mca object containing the fitted background.  If this keyword
#            is omitted then this function will call Mca.fit_background() before
#            calling fitPeaks().
#            
#         output:
#            The name of an output file to receive the ASCII printout of the
#            fit results.  This keyword is simply passed to fit_peaks_report().
#            
#         spreadsheet:
#            The name of an output file to receive the ASCII output of the
#            fit results in spreadsheet format.  This keyword is simply passed to
#            fit_peaks_report().
#            
#         append:
#            Flag indicating whether the output and spreadsheet files should be
#            appended to or overwritten.  This keyword is simply passed to
#            fit_peaks_report9).
#
#         In addition to these keywords, all keywords accepted by the
#         fit_background() function are accepted if the background keyword is
#         not present, i.e. if this function will be calling fit_background().
#
#      Outputs:
#         This function returns an Mca object which is identical to the calling
#         object, except that the data have been replaced by the peak fit.
#
#      Procedure:
#         The function does the following:
#            - Creates the Fit structure with mca->FIT_INITIALIZE() if Fit
#              was not passed as a keyword parameter.
#            - Fits the background using fit_background() if background
#              was not passed as a keyword parameter.
#            - Extracts the data from the input spectrum and the background
#              spectrum.
#            - Calls fitPeaks() with the background subtracted data.
#            - Calls fit_peaks_report()
#            - Creates a new Mca object using Mca.deepcopy and stores the output
#              of fitPeaks() in this new object with set_data().  It then
#              returns this new MCA object as the function return value.
#
#       Example:
#             mca = Mca()
#             mca.read_file('mca.001')
#             peaks = read_peaks('mypeaks.pks')
#             fit = mca.fit_peaks(peaks, bottom=6, exponent=4)
#      """
#      # If fit is not an object of type McaFit then initialize it
#      if (not isinstance(fit, McaFit)):
#         fit = McaFit(self)
#
#      # Copy the calibration parameters to fit
#      fit.update(self)
#
#      # If background is not an object of type Mca then initialize it
#      if (not isinstance(background, Mca)):
#          background = self.fit_background(**background_kw)
#
#      fit.npeaks = len(peaks)
#      background_counts = background.get_data()
#      observed_counts = self.get_data()
#      t0 = time.time()
#      [fit, peaks, fit_counts] = fitPeaks.fitPeaks(fit, peaks,
#                                      observed_counts - background_counts)
#      t1 = time.time()
#      fit_counts = fit_counts + background_counts
#      self.fit_peaks_report(fit, peaks, background, output=output, 
#                       spreadsheet=spreadsheet, append=append, time=t1-t0)
#      fit_mca = copy.copy(self)
#      fit_mca.set_data(fit_counts)
#      return([fit, peaks, fit_mca])
#
#

#   ########################################################################
#   def fit_peaks_report(self, fit, peaks, background, output='',
#                        spreadsheet=None, append=1, time=None):
#      """
#      Prints out the results from <A HREF="mca_utility_routines.html#FIT_PEAKS">FIT_PEAKS</A>
#
#      Inputs:
#         fit:
#            An McaFit object with the global fitting parameters.
#
#         peaks:
#            A list of McaPeak objects with the fit results for each peak.
#
#         See fit_peaks for more information on fit and peaks.
#
#      background:
#         An Mca object containing the fitted background spectrum.
#
#      Keywords:
#         output:
#            The name of an output file to receive the ASCII printout of the
#            fit results.  If this keyword is omitted then the output will be
#            written to stdout, i.e. the IDL output window.  If the Output file
#            already exists then the new information will (by default) be appended
#            to the file.
#
#         spreadsheet:
#            The name of an output file to receive the ASCII output of the
#            fit results in a format easily imported into a spreadsheet.  If this
#            keyword is omitted then no spreadsheet output will be generated.
#            written to stdout, i.e. the IDL output window.
#            If the spreadhseet file already exists then the new information will
#            (by default) be appended to the file.
#
#         append:
#            Set this keyword to 0 to overwrite the output and spreadsheet files
#            rather than to append to them, which is the default behavior.
#
#      Example:
#         mca = Mca(file='mca.001')
#         peaks = read_peaks('mypeaks.pks')
#         [fit, peaks, predicted] = mca.fit_peaks(peaks, fit,
#                                                 bottom=6, exponent=4)
#         mca.fit_peaks_report(fit, peaks, background, output='mca.001_out')
#      """
#
#      if (append): mode = 'a'
#      else: mode = 'w'
#      if (output == ''): out_fp = sys.stdout
#      else: out_fp = open(output, mode)
#      if (spreadsheet != None): spread_fp = open(spreadsheet, mode)
#      else: spread_fp = None
#
#      SIGMA_TO_FWHM = 2.*Numeric.sqrt(2.*Numeric.log(2.))
#
#      # Compute backgrounds
#      background_counts = background.get_data()
#      nchans = len(self.get_data())
#      for peak in peaks:
#         low = background.energy_to_channel(peak.energy - 
#                 2.*peak.fwhm / SIGMA_TO_FWHM)
#         low = min(max(low, 0), nchans-3)
#         hi  = background.energy_to_channel(peak.energy + 
#                 2.*peak.fwhm / SIGMA_TO_FWHM)
#         hi = min(max(hi, low+1), nchans-1)
#         peak.bgd = Numeric.sum(background_counts[low:hi+1])
#
#      if (out_fp != None):
#         out_fp.write('\n')
#         out_fp.write('*******************************************************\n')
#         out_fp.write( '    Fit of ' + self.name + '\n')
#         out_fp.write('\n')
#         elapsed = self.get_elapsed()
#         out_fp.write('Real time (seconds):            ' +
#                        ('%10.2f' % elapsed.real_time) + '\n')
#         out_fp.write('Live time (seconds):            ' +
#                        ('%10.2f' % elapsed.live_time) + '\n')
#         out_fp.write('Initial FWHM offset, slope:     ' +
#                        ('%10.6f' % fit.initial_fwhm_offset) + 
#                        ('%10.6f' % fit.initial_fwhm_slope) + '\n')
#         out_fp.write('Optimized FWHM offset, slope:   ' +
#                        ('%10.6f' % fit.fwhm_offset) +
#                        ('%10.6f' % fit.fwhm_slope) +  '\n')
#         out_fp.write('Initial energy offset, slope:   ' +
#                        ('%10.6f' % fit.initial_energy_offset) +
#                        ('%10.6f' % fit.initial_energy_slope) + '\n')
#         out_fp.write('Optimized energy offset, slope: ' +
#                        ('%10.6f' % fit.energy_offset) +
#                        ('%10.6f' % fit.energy_slope) + '\n')
#         out_fp.write('# Iterations, function evals:   ' +
#                        ('%10d' % fit.n_iter) +
#                        ('%10d' % fit.n_eval) + '\n')
#         out_fp.write('Chi squared:                    ' +
#                        ('%.6g' % fit.chisqr) + '\n')
#         out_fp.write('Status code:                    ' +
#                        ('%d' % fit.status) + '\n')
#         if (fit.status <= 0):
#            out_fp.write('Error message:               ' +
#                           fit.err_string)
#         if (time != None):
#            out_fp.write('Time to fit:                 ' +
#                        ('%.3f' % time) + '\n')
#             
#         out_fp.write('\n')
#         out_fp.write('        Peak       Energy    FWHM      Area     ' + 
#                'Background   Area/MDL   Area/Bkg\n')
#         out_fp.write('\n')
#         for peak in peaks:
#            if (peak.energy_flag == 0):  esym=' '
#            else: esym='*'
#            if (peak.fwhm_flag == 0):    fsym=' '
#            else: fsym='*'
#            if (peak.ampl_factor == 0.): asym=' '
#            else: asym='*'
#            out_fp.write(
#                  ('%15s' % peak.label) +
#                  ('%10.3f%1s' % (peak.energy, esym)) +
#                  ('%10.4f%1s' % (peak.fwhm, fsym)) +
#                  ('%10.1f%1s' % (peak.area, asym)) +
#                  ('%10.1f' % peak.bgd) +
#                  ('%10.1f' % (peak.area/
#                               max((3.*Numeric.sqrt(peak.bgd)), 1.0))) +
#                  ('%10.1f' % (peak.area/max(peak.bgd, 1.0)))+'\n')
#
#      if (spread_fp != None):
#         el = ('%10.2f' % elapsed.live_time)
#         er = ('%10.2f' % elapsed.real_time)
#         spread_fp.write(self.name+'#Labels#Live time#Real time#')
#         for peak in peaks:
#            spread_fp.write(peak.label+'#')
#         spread_fp.write('\n')
#         spread_fp.write(self.name+'#Energy#'+el+'#'+er+'#')
#         for peak in peaks:
#            spread_fp.write(('%10.3f' % peak.energy)+'#')
#         spread_fp.write('\n')
#         spread_fp.write(self.name+'#FWHM#'+el+'#'+er+'#')
#         for peak in peaks:
#            spread_fp.write(('%10.4f' % peak.fwhm)+'#')
#         spread_fp.write('\n')
#         spread_fp.write(self.name+'#Area#'+el+'#'+er+'#')
#         for peak in peaks:
#            spread_fp.write(('%10.1f' % peak.area)+'#')
#         spread_fp.write('\n')
#         spread_fp.write(self.name+'#Background#'+el+'#'+er+'#')
#         for peak in peaks:
#            spread_fp.write(('%10.1f' % peak.bgd)+'#')
#         spread_fp.write('\n')
#         spread_fp.close()

#######################################################################
def write_ascii_file(file, data, calibration, elapsed, presets, rois,
                     environment):
   """
   Writes Mca or Med data to a disk file.  The file 
   format is a tagged ASCII format.  The file contains the information 
   from the Mca object which it makes sense to store permanently, but 
   does not contain all of the internal state information for the Mca.  
   Files written with this routine can be read with read_ascii_file(), which
   is called by Mca.read_file() if the netcdf flag is 0.

   This procedure is typically not called directly, but is called
   by Mca.write_file if the netcdf=1 keyword is not used.

   This function can be used for writing for Mca objects, in which case
   each input parameter is an object, such as McaElapsed, etc.
   It can also be used for writing Med objects, in which case each input
   parameter is a list.
   
   If the rank of data is 2 then this is an Med, and the number of detectors
   is the first dimension of data

   Inputs:
      file:
         The name of the disk file to write.
         
      data:
         The data to write.  Either 1-D array or list of 1-D arrays.

      calibration:
         An object of type McaCalibration, or a list of such objects.

      elapsed:
         An object of type McaElapsed, or a list of such objects.

      presets:
         An object of type McaPresets, or a list of such objects.

      rois:
         A list of McaROI objects, or a list of lists of such objects.
      
      environment:
         A list of McaEnvironment objects, or a list of lists of such objects.
   """
   if (Numeric.rank(data) == 2):
      n_det = len(data)
   else:
      n_det = 1
   fformat = '%f ' * n_det
   eformat = '%e ' * n_det
   iformat = '%d ' * n_det
   sformat = '%s ' * n_det
   if (n_det == 1):
      # For convenience we convert all attributes to lists
      data = [data]
      rois = [rois]
      calibration = [calibration]
      presets = [presets]
      elapsed = [elapsed]
   nchans = len(data[0])
   start_time = elapsed[0].start_time

   fp = open(file, 'w')
   fp.write('VERSION:    '+'3.1'+'\n')
   fp.write('ELEMENTS:   '+str(n_det)+'\n')
   fp.write('DATE:       '+str(start_time)+'\n')
   fp.write('CHANNELS:   '+str(nchans)+'\n')

   nrois = []
   for roi in rois:
      nrois.append(len(roi))
   fp.write('ROIS:       '+(iformat % tuple(nrois))+'\n')
   real_time=[]; live_time=[]
   for e in elapsed:
      real_time.append(e.real_time)
      live_time.append(e.live_time)
   fp.write('REAL_TIME:  '+(fformat % tuple(real_time))+'\n')
   fp.write('LIVE_TIME:  '+(fformat % tuple(live_time))+'\n')
   offset=[]; slope=[]; quad=[]; two_theta=[]
   for c in calibration:
      offset.append(c.offset)
      slope.append(c.slope)
      quad.append(c.quad)
      two_theta.append(c.two_theta)
   fp.write('CAL_OFFSET: '+(eformat % tuple(offset))+'\n')
   fp.write('CAL_SLOPE: '+(eformat % tuple(slope))+'\n')
   fp.write('CAL_QUAD: '+(eformat % tuple(quad))+'\n')
   fp.write('TWO_THETA: '+(fformat % tuple(two_theta))+'\n')

   for i in range(max(nrois)):
      num = str(i)
      left=[]; right=[]; label=[]
      for d in range(n_det):
         if (i < nrois[d]):
            left.append(rois[d][i].left)
            right.append(rois[d][i].right)
            label.append(rois[d][i].label + '&')
         else:
            left.append(0)
            right.append(0)
            label.append(' &')
      fp.write('ROI_'+num+'_LEFT:  '+(iformat % tuple(left))+'\n')
      fp.write('ROI_'+num+'_RIGHT:  '+(iformat % tuple(right))+'\n')
      fp.write('ROI_'+num+'_LABEL:  '+(sformat % tuple(label))+'\n')
   for e in environment:
     fp.write('ENVIRONMENT: '       + str(e.name) +
                              '="'  + str(e.value) +
                              '" (' + str(e.description) + ')\n')
   fp.write('DATA: \n')
   counts = Numeric.zeros(n_det)
   for i in range(nchans):
      for d in range(n_det):
         counts[d]=data[d][i]
      fp.write((iformat % tuple(counts))+'\n')
   fp.close()

########################################################################
def read_ascii_file(file):
   """
   Reads a disk file.  The file format is a tagged ASCII format.
   The file contains the information from the Mca object which it makes sense
   to store permanently, but does not contain all of the internal state
   information for the Mca.  This procedure reads files written with
   write_ascii_file().

   Inputs:
      file:
         The name of the disk file to read.
         
   Outputs:
      Returns a dictionary of the following type:
      'n_detectors': int,
      'calibration': [McaCalibration()],
      'elapsed':     [McaElapsed()],
      'rois':        [[McaROI()]]
      'data':        [Numeric.array]
      'environment': [[McaEnvironment()]]
      
   Example:
      m = read_ascii_file('mca.001')
      m['elapsed'][0].real_time
   """
   fp = open(file, 'r')
   line = ''
   start_time = ''
   data = None
 
   environment = []
   n_detectors = 1  # Assume single element data
   elapsed = [McaElapsed()]
   calibration = [McaCalibration()]
   rois = [[]]
   while(1):
      line = fp.readline()
      if (line == ''): break
      pos = line.find(' ') #pos = string.find(line, ' ')
      if (pos == -1): pos = len(line)
      tag = line[0:pos]
      value = str.strip(line[pos:])#line[pos:].strip #value = string.strip(line[pos:])
      values = str.split(value) #values = string.split(value)
      if (tag == 'VERSION:'):
          pass
      elif (tag == 'DATE:'):  
         start_time = value
      elif (tag == 'ELEMENTS:'):
         n_detectors  = int(value)
         for det in range(1, n_detectors):
             elapsed.append(McaElapsed())
             calibration.append(McaCalibration())
             rois.append([])
      elif (tag == 'CHANNELS:'):
         nchans = int(value)
      elif (tag == 'ROIS:'):
         nrois = []
         for d in range(n_detectors):
            nrois.append(int(values[d]))
         max_rois = max(nrois)
         for d in range(n_detectors):
            for r in range(nrois[d]):
               rois[d].append(McaROI())
      elif (tag == 'REAL_TIME:'):
         for d in range(n_detectors):
            elapsed[d].start_time = start_time
            elapsed[d].real_time = float(values[d])
      elif (tag == 'LIVE_TIME:'):  
         for d in range(n_detectors):
            elapsed[d].live_time = float(values[d])
      elif (tag == 'CAL_OFFSET:'):
         for d in range(n_detectors):
            calibration[d].offset = float(values[d])
      elif (tag == 'CAL_SLOPE:'):
         for d in range(n_detectors):
            calibration[d].slope = float(values[d])
      elif (tag == 'CAL_QUAD:'):  
         for d in range(n_detectors):
            calibration[d].quad = float(values[d])
      elif (tag == 'TWO_THETA:'):
         for d in range(n_detectors):
            calibration[d].two_theta = float(values[d])
      elif (tag == 'ENVIRONMENT:'):
         env = McaEnvironment()
         p1 = str.find(value, '=')
         env.name = value[0:p1]
         p2 = str.find(value[p1+2:], '"')
         env.value = value[p1+2: p1+2+p2]
         env.description = value[p1+2+p2+3:-1]
         environment.append(env)
      elif (tag == 'DATA:'):
         data = []
         for d in range(n_detectors):
            data.append(Numeric.zeros(nchans, 'i'))
         for chan in range(nchans):
            line = fp.readline()
            counts = str.split(line)
            for d in range(n_detectors):
               data[d][chan]=int(counts[d])
      else:
         for i in range(max_rois):
             roi = 'ROI_'+str(i)+'_'
             if (tag == roi+'LEFT:'):
                for d in range(n_detectors):
                   if (i < nrois[d]):
                       rois[d][i].left = int(values[d])
                break
             elif (tag == roi+'RIGHT:'):
                for d in range(n_detectors):
                   if (i < nrois[d]):
                      rois[d][i].right = int(values[d])
                break
             elif (tag == roi+'LABEL:'):
                labels = str.split(value, '&')
                for d in range(n_detectors):
                   if (i < nrois[d]):
                      rois[d][i].label = str.strip(labels[d])
                break
         else:
            print('Unknown tag = '+tag+' in file: ' + file + '.')

   # Make sure DATA array is defined, else this was not a valid data file
   if (data == None): print('Not a valid data file: ' + file + '.')
   fp.close()
   # Built dictionary to return
   r = {}
   r['n_detectors'] = n_detectors
   r['calibration'] = calibration
   r['elapsed'] = elapsed
   r['rois'] = rois
   r['data'] = data
   r['environment'] = environment
   return r

########################################################################
#def read_peaks(file):
#   """
#   Reads a disk file into an array of structures of type 
#   McaPeak.  This routine is typically called before calling fit_peaks.
#   The routine also returns a structure of type McaBackground.  This structure
#   may or may not actually be defined in the file (older peaks files lacked it),
#   but a reasonable default value will always be returned.
#
#   Inputs:
#      file:  
#         The name of a disk file containing the peak definitions.
#         
#   Outputs:
#        This function returns a dictionary:
#        'peaks': [McaPeak]          # A list of McaPeak objects
#        'background': McaBackground # An McaBackground object
#        
#   The format of the disk file is as follows:
#      - Lines containing the parameters for the background (fields in
#        McaBackground class) have the following format:
#           Background_exponent, 4
#           Background_top_width, 0
#           Background_bottom_width, 4
#           Background_tangent, 0
#           Background_compress, 8
#      - There is one line in the file for each peak
#           - Each line consists of the following fields, separated by commas:
#             energy, energy_flag, fwhm, fwhm_flag, ampl_factor, label
#           - All fields except the first, "energy", are optional and default
#             values of 0 or blank.
#           - Field descriptions
#               - energy:       This field can either be an energy in keV or a
#                               string which can be parsed by Xrf.lookup_xrf_line()
#                               or Xrf.lookup_gamma_line()
#                               The energy in keV of the peak is put in the 
#                               .energy value for the peak.
#               - energy_flag:  The .energy_flag value for the peak. 0 or 1.
#               - fwhm:         The .initial_fwhm value for the peak.  This 
#                               can be 0 if the .fwhm_flag is 0.
#               - fwhm_flag:    The .fwhm_flag value for the peak. 0, 1 or 2.
#               - ampl_factor:  The .ampl_factor for the peak.
#               - label:        The .label string for the peak.  If the energy
#                               field is a string, and the label field is blank
#                               then the energy string will be put in .label.
#
#   The following is an example peak file. Most peaks in this example use 
#   the default (0) values for the energy_flag, fwhm_flag and
#   amplitude_factor fields.
#      4.660,,,,,Fe escape  ! This peak uses a numeric energy value
#      Fe Ka                ! This peak uses a string energy value
#      Fe Kb,,,,.5          ! Fe Kb is constrained to be 0.5 of Fe Ka
#      Ni Ka                ! These peaks energies are found with
#      Ni Kb                !  LOOKUP_XRF_LINE
#      Co57 G2              ! These peak energies are found with
#      Cd109 G1             !  LOOKUP_GAMMA_LINE
#      15.9,1,.3,1,,Diffraction ! Fit both the energy and fwhm of this peak
#      17.443,1,,,,Unknown  ! Fit the energy, but not fwhm, of this peak
#   """
#
#   fp = open(file, 'r')
#   lines = fp.readlines()
#   fp.close()
#   peaks = []
#   background = McaBackground()
#   for s in lines:
#      s = s.strip()
#      q = s.find(',')
#      if (q >= 0): 
#         label = s[0:q].upper()
#         if (label == "BACKGROUND_EXPONENT"):
#            background.exponent = int(float(s[q+1:]))
#         elif (label == "BACKGROUND_TOP_WIDTH"):
#            background.top_width = float(s[q+1:])
#         elif (label == "BACKGROUND_BOTTOM_WIDTH"):
#            background.bottom_width = float(s[q+1:])
#         elif (label == "BACKGROUND_TANGENT"):
#            background.tangent = int(s[q+1:])
#         elif (label == "BACKGROUND_COMPRESS"):
#            background.compress = int(s[q+1:])
#         else:
#            peaks.append(parse_peak(s))
#      else:
#         peaks.append(parse_peak(s))
#   return {'peaks': peaks, 'background': background}
#
#########################################################################
#def parse_peak(text):
#   """ Private function """
#   peak = McaPeak()
#   params = text.split(',')
#   n_params = len(params)
#   for i in range(n_params):
#      params[i] = params[i].strip()
#   if (n_params == 0): return peak
#
#   try:
#      peak.initial_energy = float(params[0])
#   except:
#      peak.label = params[0]
#      e = Xrf.lookup_xrf_line(params[0])
#      if (e == None):
#         e = Xrf.lookup_gamma_line(params[0])
#      if (e == None): e = 0.
#      peak.initial_energy = e
#   if (n_params == 1): return peak
#
#   if (len(params[1]) > 0):
#      peak.energy_flag = int(float(params[1]))
#   if (n_params == 2): return peak
#
#   if (len(params[2]) > 0):
#      peak.intial_fwhm = float(params[2])
#   if (n_params == 3): return peak
#
#   if (len(params[3]) > 0):
#      peak.fwhm_flag = int(float(params[3]))
#   if (n_params == 4): return peak
#
#   if (len(params[4]) > 0):
#      peak.ampl_factor = float(params[4])
#   if (n_params == 5): return peak
#
#   if (len(params[5]) > 0):
#      peak.label = params[5]
#   return peak

########################################################################
def write_peaks(file, peaks, background=None):
   """
   Writes a list of obejcts of type McaPeak to a disk file.
   If the background parameter is present it also writes the background
   structure to the file.
   
   Inputs:
      file:
         The name of a disk file to be written ;
      peaks:
         A list of McaPeak objects
         
   Keywords:
      background:
         An object of type McaBackground
         
   Example:
      r = read_peaks('my_peaks.pks')
      peaks = r['peaks']
      peaks[1].initial_energy = 6.4
      write_peaks('mypeaks.pks', peaks)
   """
   lines = []
   if (background != None):
      lines.append('Background_exponent,'     + str(background.exponent)+'\n')
      lines.append('Background_top_width,'    + str(background.top_width)+'\n')
      lines.append('Background_bottom_width,' + str(background.bottom_width)+'\n')
      lines.append('Background_tangent,'      + str(background.tangent)+'\n')
      lines.append('Background_compress,'     + str(background.compress)+'\n')
   for peak in peaks:
      lines.append(str(peak.initial_energy) + ', ' + 
                   str(peak.energy_flag)    + ', ' + 
                   str(peak.initial_fwhm)   + ', ' + 
                   str(peak.fwhm_flag)      + ', ' + 
                   str(peak.ampl_factor)    + ', ' + 
                   str(peak.label) + '\n')
   fp = open(file, 'w')
   fp.writelines(lines)
   fp.close()




__all__ = ['XRD_FitPattern', 'XRD_FitSubpattern', 'DioptasFunctions', 'GSASIIFunctions', 'MedFunctions',
           'PeakFunctions', 'Plot85Functions',  'WriteMultiFit', 'WritePolydefixED', 'XRD_FitPattern',
           'XRD_FitSubpattern', 'ImageMetaData', 'Mca', 'Med', 'generate_inputs']

from . import XRD_FitPattern
from . import XRD_FitSubpattern
from . import PeakFunctions
from . import DioptasFunctions
from . import GSASIIFunctions
from . import MedFunctions
from . import Plot85Functions
from . import WriteMultiFit
from . import WritePolydefix
from . import WritePolydefixED
from . import WriteDifferentialStrain
from . import Mca
from . import Med


def generate_inputs():

    try:
        fp = open('Example_MultiPeak_Inputs.py', 'r')
        fp.close()
        raise UserWarning('Example inputs already exist')
    except FileNotFoundError:
        pass

    try:
        fp = open('Example_MultiPeak_Inputs.py', 'w')

        fp.write("# Input parameters for BCC1_2GPa_100s_001.\n#\n\n\n")

        fp.write("# properties of the data files.\n")
        fp.write("datafile_directory = './Example1-Fe/'\n")
        fp.write("datafile_Basename  = 'BCC1_2GPa_10s_001_'\n")
        fp.write("datafile_Ending    = '.tif'\n")
        fp.write("datafile_StartNum  = 1\n")
        fp.write("datafile_EndNum    = 10\n")
        fp.write("datafile_NumDigit  = 5\n\n")

        fp.write("# Calibration and masking.\n")
        fp.write("Calib_type   = 'Dioptas'\n")
        fp.write("Calib_detector = 'Pilatus1M'\n")
        fp.write("Calib_data     = datafile_directory + 'CeO2_Pil207_E30_2Nov2016_001.tif'\n")
        fp.write("Calib_param    = datafile_directory + 'CeO2_cal_Dioptas.poni'\n")
        fp.write("Calib_mask     = datafile_directory + 'DiffractionMask_Dioptas.mask'\n")
        fp.write("Calib_pixels = 172\n\n")

        fp.write("# Fitting properties for peaks.\n")
        fp.write("fit_orders = [\n")
        fp.write("       {\n")
        fp.write("         'background': [0,0],\n")
        fp.write("         'peak': [{\n")
        fp.write("             'phase': 'Other',\n")
        fp.write("             'hkl': '000',\n")
        fp.write("             'd-space': 3,\n")
        fp.write("             'height': 1,\n")
        fp.write("             'profile': 0,\n")
        fp.write("             #'profile_fixed': 1,\n")
        fp.write("             'width': 0,\n")
        fp.write("             'symmetry': 2\n")
        fp.write("           },{\n")
        fp.write("             'phase': 'BCC-Fe',\n")
        fp.write("             'hkl': 110,\n")
        fp.write("             'd-space': 3,\n")
        fp.write("             'height': 16,\n")
        fp.write("             'profile': 0,\n")
        fp.write("             #'profile_fixed': 1,\n")
        fp.write("             'width': 0,\n")
        fp.write("             'symmetry': 2\n")
        fp.write("           }],\n")
        fp.write("         'range': [[10.8, 11.67]],\n")
        fp.write("         'PeakPositionSelection': [[1,-170.5,11.19],[1,-150.5,11.19],[1,-120.5,11.19],[1,-58.997,11.19],[1,59.289,11.19],[1,23.187,11.19],[1,23.212,11.19], [1,53.158,11.19], [1,73.246,11.19], [1,93.246,11.19], [1,123.246,11.19], [2,-170, 11.54], [2,-150, 11.54], [2,-110, 11.54], [2,-90, 11.54], [2,-45, 11.54], [2,-0, 11.54], [2,45, 11.54], [2,90, 11.54], [2,120, 11.54], [2,150, 11.54], [2,179, 11.54]]\n")
        fp.write("},\n")
        fp.write("       {\n")
        fp.write("         'background': [0,0],\n")
        fp.write("         'peak': [{\n")
        fp.write("             'phase': 'BCC-Fe',\n")
        fp.write("             'hkl': 200,\n")
        fp.write("             'd-space': 3,\n")
        fp.write("             'height': 8,\n")
        fp.write("             'profile': 0,\n")
        fp.write("             #'profile_fixed': 1,\n")
        fp.write("             'width': 0,\n")
        fp.write("             'symmetry': 2\n")
        fp.write("           }],\n")
        fp.write("         'range': [[16.1,16.6]]\n")
        fp.write("       },\n")
        fp.write("       {\n")
        fp.write("         'background': [0,0],\n")
        fp.write("         'peak': [{\n")
        fp.write("             'phase': 'BCC-Fe',\n")
        fp.write("             'hkl': 211,\n")
        fp.write("             'd-space': 3,\n")
        fp.write("             'height': 8,\n")
        fp.write("             'profile': 0,\n")
        fp.write("             #'profile_fixed': 1,\n")
        fp.write("             'width': 0,\n")
        fp.write("             'symmetry': 2\n")
        fp.write("           }],\n")
        fp.write("         'range': [[19.8, 20.27]]\n")
        fp.write("       },\n")
        fp.write("      {\n")
        fp.write("          'background': [1],\n")
        fp.write("          'peak': [{\n")
        fp.write("             'phase': 'BCC-Fe',\n")
        fp.write("             'hkl': 220,\n")
        fp.write("              'd-space': 3,\n")
        fp.write("              'height': 8,\n")
        fp.write("              'profile': 0,\n")
        fp.write("              'profile_fixed': 0.5,\n")
        fp.write("              'width': 0,\n")
        fp.write("              'symmetry': 2\n")
        fp.write("            }],\n")
        fp.write("#          'PeakPositionSelection': [[1,-120.5,23.226],[1,-58.997,23.186],[1,59.289,23.208],[1,23.187,120.893],[1,23.212,6.367], [1,23.158,90.025], [1,23.246,25.729]]\n")
        fp.write("          'range': [[23.1, 23.5]]\n")
        fp.write("        },\n")
        fp.write("       {\n")
        fp.write("         'background': [1,1],\n")
        fp.write("         'peak': [{\n")
        fp.write("             'phase': 'BCC-Fe',\n")
        fp.write("             'hkl': 310,\n")
        fp.write("             'd-space': 2,\n")
        fp.write("             'height': 1,\n")
        fp.write("             'profile': 0,\n")
        fp.write("             'profile_fixed': 0.5,\n")
        fp.write("             'width': 0,\n")
        fp.write("             'symmetry': 2\n")
        fp.write("           }],\n")
        fp.write("         'range': [[25.7, 26.4]]\n")
        fp.write("       },\n")
        fp.write("    ]\n\n")

        fp.write("#Number of bins for initial fitting.\n")
        fp.write("AziBins = 90\n\n")


        fp.write("#Output settings\n")
        fp.write("Output_directory   = './'\n")
        fp.write("Output_type        = 'MultiFit'\n")
        fp.write("Output_NumAziWrite = 90\n\n")

        fp.close()


    except PermissionError:
        print('Write permission required')


#!/usr/bin/env python
import os
import sys
from PIL import Image
from PIL.ExifTags import TAGS

Im = sys.argv[1]

print 'Image File: ', Im

with Image.open(Im) as img:
    meta_dict = {TAGS[key] : img.tag[key] for key in img.tag.iterkeys()}

for (tag,value) in Image.open(Im)._getexif().iteritems():
        print '%s = %s' % (TAGS.get(tag), value)


#!/usr/bin/env python

__all__ = ['FitSubpattern']

# FPF_XRD_FitSubpattern
# Script fits subset of the data with peaks of pre-defined Fourier orders

import os
import sys
import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
from lmfit import Parameters, Model
from lmfit.model import save_modelresult#, load_modelresult
import importlib

import cpf.PeakFunctions as ff

np.set_printoptions(threshold=sys.maxsize)


# FIX ME;
# The script should not need to know what the limits are - it should only see the data that it needs to fit.
def peak_string(orders, fname=False):
    """
    :param orders: list of peak orders which should include peak names/hkls
    :return: string listing peak names
    """
    #FIX ME: this will be a data type function so should probably more somewhere else in the end. 
    p_str = ''
    
    for x in range(len(orders['peak'])):
        if 'phase' in orders['peak'][x]:
            p_str = p_str + orders['peak'][x]['phase']
        else:
            p_str = p_str + "Peak"
        if fname is False:
            p_str = p_str + " ("
        if 'hkl' in orders['peak'][x]:
            p_str = p_str + str(orders['peak'][x]['hkl'])
        else:
            p_str = p_str + str(x+1)
        if fname is False:
            p_str = p_str + ")"
                
        if x < len(orders['peak']) - 1 and len(orders['peak']) > 1:
            if fname is False:
                p_str = p_str + " & "
            else:
                p_str = p_str + '_'

    return p_str


def run_initial_fit(azi, y_data, y_data_errs, params, param_str, comp, order, extras, method='leastsq', symm=1):
    """
    :param azi: x-data arr float
    :param y_data: y_data arr float
    :param y_data_errs: y_data error arr float
    :param params: lmfit Parameter object
    :param param_str: base string to identify components str
    :param comp: string to add to base string to identify components str
    :param order: determines number of Fourier coefficients int
    :param method: lmfit method to use str
    :param extras: list of min, max and expression for bounds on parameters
    :return: lmfit Parameter class
    """
    min, max, expr = extras
    params = ff.initiate_params(params, param_str, comp, order, max=max, min=min, expr=expr)
    params = ff.unvary_params(params, param_str, comp)  # set other parameters to not vary
    fout = ff.Fourier_fit(azimu=azi, ydata=np.array(y_data), param=params, errs=np.array(y_data_errs),
                          param_str=param_str + '_' + comp, symm=symm, fit_method=method)
    params = fout.params

    return params


def initial_component_fits(master_params, param_str, data_arrays, comp_lists, order_list, symm, limits, pfixed=None):
    """
    Perform initial Fourier fits to each component data in turn
    :param master_params: lmfit Parameter class object
    :param param_str: start string to label parameters
    :param data_arrays: list of data arrays for use in component fits
    :param comp_lists: list of lists for parameters
    :param order_list: list of orders for components
    :param symm : number definiing the symmetry
    :param limits: array of limits
    :param pfixed=None: either False or true
    :return: list of lists of each component with coefficients and errors
    """

    newAziChunks, newd0, newd0Err, newHall, newHallErr, newWall, newWallErr, newPall, newPallErr = data_arrays
    dfour, hfour, wfour, pfour = comp_lists
    d0_order, h_order, w_order, p_order = order_list
    d_lims, h_lims, w_lims, p_lims = limits

    # initiate symmetry
    comp = 's'
    master_params = ff.initiate_params(master_params, param_str, comp, 0, max=10, min=1, value=symm)

    # FIX ME: need to propagate method properly
    comp = 'd'
    master_params = run_initial_fit(azi=np.array(newAziChunks), y_data=np.array(newd0),
                                    y_data_errs=np.array(newd0Err), params=master_params, param_str=param_str,
                                    comp=comp, order=d0_order, extras=d_lims, method='leastsq', symm=1)
    dfour.append(ff.gather_paramerrs_to_list(master_params, param_str, comp))

    comp = 'h'
    master_params = run_initial_fit(azi=np.array(newAziChunks), y_data=np.array(newHall),
                                    y_data_errs=np.array(newHallErr), params=master_params, param_str=param_str,
                                    comp=comp, order=h_order, extras=h_lims, method='leastsq', symm=symm)
    hfour.append(ff.gather_paramerrs_to_list(master_params, param_str, comp))

    comp = 'w'
    master_params = run_initial_fit(azi=np.array(newAziChunks), y_data=np.array(newWall),
                                    y_data_errs=np.array(newWallErr), params=master_params, param_str=param_str,
                                    comp=comp, order=w_order, extras=w_lims, method='leastsq', symm=symm)
    wfour.append(ff.gather_paramerrs_to_list(master_params, param_str, comp))

    # FIX ME: seems to occur regardless of pfixed as checked in other places, is this correct? SAH: NO. Now Fixed.
    comp = 'p'
    if pfixed is False:
        master_params = run_initial_fit(azi=np.array(newAziChunks), y_data=np.array(newPall),
                                        y_data_errs=np.array(newPallErr), params=master_params, param_str=param_str,
                                        comp=comp, order=p_order, extras=p_lims, method='leastsq', symm=symm)
        #pfour.append(ff.gather_paramerrs_to_list(master_params, param_str, comp))
    else:
        master_params = ff.initiate_params(master_params, param_str=param_str, comp=comp, value=pfixed)    
    pfour.append(ff.gather_paramerrs_to_list(master_params, param_str, comp))
    # master_params.pretty_print()

    #print('Initial fit output parms')
    #print(dfour, hfour, wfour, pfour)
    #master_params.pretty_print()

    return master_params, dfour, hfour, wfour, pfour


def update_component_fits(master_params, intens, azimu, twotheta, param_str, comp_lists, conversion_factor, bgguess,
                          Guesses, pfixed=None):
    """
    Perform component fits to data
    :param master_params: lmfit Parameter class object
    :param intens: intensity data array
    :param azimu: data array
    :param twotheta: data array
    :param param_str: start string to label parameters
    :param comp_lists: list of lists for parameters
    :param conversion_factor: dict inputs for conversion
    :param bgguess: initial values or background dict
    :param Guess: initial values for peak components dict
    :param pfixed: fix p or not
    :return: list of lists with component coefficients
    """

    dfour, hfour, wfour, pfour = comp_lists

    comp = 'd'
    master_params = ff.unvary_params(master_params, param_str, comp)  # set other parameters to not vary
    master_params = ff.vary_params(master_params, param_str, comp)  # set these parameters to vary
    # master_params.pretty_print()
    out = ff.FitModel(intens.flatten(), twotheta.flatten(), azimu.flatten(),
                      num_peaks=len(Guesses['peak']), nterms_back=len(bgguess), Conv=conversion_factor,
                      fixed=pfixed, method=None, weights=None, params=master_params)
    # master_params = out.params
    dfour.append(ff.gather_paramerrs_to_list(master_params, param_str, comp))

    comp = 'h'
    master_params = ff.unvary_params(master_params, param_str, comp)  # set other parameters to not vary
    master_params = ff.vary_params(master_params, param_str, comp)  # set these parameters to vary
    out = ff.FitModel(intens.flatten(), twotheta.flatten(), azimu.flatten(),
                      num_peaks=len(Guesses['peak']), nterms_back=len(bgguess), Conv=conversion_factor,
                      fixed=pfixed, method=None, weights=None, params=master_params)
    master_params = out.params
    hfour.append(ff.gather_paramerrs_to_list(master_params, param_str, comp))

    comp = 'w'
    master_params = ff.unvary_params(master_params, param_str, comp)  # set other parameters to not vary
    master_params = ff.vary_params(master_params, param_str, comp)  # set these parameters to vary
    out = ff.FitModel(intens.flatten(), twotheta.flatten(), azimu.flatten(),
                      num_peaks=len(Guesses['peak']), nterms_back=len(bgguess), Conv=conversion_factor,
                      fixed=pfixed, method=None, weights=None, params=master_params)
    master_params = out.params
    wfour.append(ff.gather_paramerrs_to_list(master_params, param_str, comp))

    comp = 'p'
    if pfixed is False:
        # if 'profile_fixed' not in order_peak[0]: # orders['peak'][0]:
        master_params = ff.unvary_params(master_params, param_str, comp)  # set other parameters to not vary
        master_params = ff.vary_params(master_params, param_str, comp)  # set these parameters to vary
        out = ff.FitModel(intens.flatten(), twotheta.flatten(), azimu.flatten(),
                          num_peaks=len(Guesses['peak']), nterms_back=len(bgguess),
                          Conv=conversion_factor,
                          fixed=pfixed, method=None, weights=None, params=master_params)
        master_params = out.params
    pfour.append(ff.gather_paramerrs_to_list(master_params, param_str, comp))

    return master_params, dfour, hfour, wfour, pfour


def FitSubpattern(TwoThetaAndDspacings, azimu, intens, orders=None, PreviousParams=None,
                  DetFuncs='DioptasFunctions', SaveFit=None, debug=False, refine=True, iterations=1, fnam=None):
    """
    Perform the various fitting stages to the data
    :param TwoThetaAndDspacings:
    :param azimu:
    :param intens:
    :param orders:
    :param PreviousParams:
    :param DetFuncs:
    :param SaveFit:
    :param debug:
    :return:
    """
    # load detector functions to use where needed. Mostly for conversion of tth to d-spacing.
    # This is loaded from detector functions file to avoid making any assumptions about the conversion between
    # values (if indeed any is required).
    # print(DetFuncs)
    # if DetFuncs not in sys.modules:
    det = importlib.import_module(DetFuncs)
    # det = __import__(DetFuncs)

    # sort inputs

    # two theta and d-spacing are both 'x' dimensions. therefore import them as single array.
    # split them up here if needed. 
    if np.shape(TwoThetaAndDspacings)[0] == 3:
        twotheta = TwoThetaAndDspacings[0]
        dspace = TwoThetaAndDspacings[1]
        conversion_factor = TwoThetaAndDspacings[2]
    else:
        twotheta = TwoThetaAndDspacings
        # FIX ME: need to set a value for 'conversion_factor' in here and propagate its none-use through the code.

    if debug: #get limits for chunk fit plots.
        tthrange = [np.min(twotheta.flatten()), np.max(twotheta.flatten())]

    # set data type for the intensity data.
    # This is needed for saving the fits using save_modelresult/ load_modelresult. 
    # load_modelresult fails if the data is a masked integer array.
    if isinstance(intens, int) or isinstance(intens, np.int32):
        intens = float(intens)
    elif isinstance(intens, np.int64):
        intens = np.float64(intens)
    #FIX ME: this doesnt work. the data type is got by print(intens.dtype)

    # FIX ME: need to match orders of arrays to previous numbers.
    # if no parameters
    #     get orders of arrays
    # else if no orders
    #     use parameters as given
    # else
    #     change the size of the parameter arrays to match orders

    # define the number of peaks to be fitted
    peeks = len(orders['peak'])
    # DMF: FIX ME: orders used before checked?

    if orders:
        # print orders
        #total = orders['AziBins']
        backg_type = 'order'
        bg_order = orders['background']
        # print(bg_order, 'order')

        #        #FIX ME: assumes a fixed number of peaks.
        #        if 'symmetry' in orders['peak'][0]:
        #            symmetry = orders['peak'][0]['symmetry']
        #        else:
        #            symmetry = 1

        # print bg_order
        backg = []
        '''
        # removed because if using orders it should be list of orders i.e. number in list = num. orders needed
        for j in range(len(bg_order)):
            backg.append(np.zeros(len(bg_order) * 2 + 1))
        '''
        for j in bg_order:
            backg.append(np.zeros(j * 2 + 1))
        # print(backg, 'backg')
        
        
        for y in range(peeks):
            
            # force profile orders to match the fixed orders
            # FIX ME: should this generate an error?
            if 'profile_fixed' in orders['peak'][y]:        
                if not ff.Fourier_order(orders['peak'][y]['profile_fixed']) == orders['peak'][y]['profile']:
                    print('Peak ' + str(y) + ': The order of the profile does not match that of the fixed profile. Changing profile to match fixed profile.')
                    orders['peak'][y]['profile'] = ff.Fourier_order(orders['peak'][y]['profile_fixed'])
        
                #make sure the fixed profile is a list
                if not isinstance(orders['peak'][y]['profile_fixed'], list):
                    orders['peak'][y]['profile_fixed'] = [orders['peak'][y]['profile_fixed']]


    if PreviousParams and orders:
        # need to check sizes of both arrays here.
        # orders takes precedence over PreviousParams so we add or remove coefficients as necessary

        for y in range(peeks):

            # loop over parameters
            for x in range(4):
                # DMF: Why not just loop for Parm in orders['peak'][y]?

                # only 4 parameters describe each peak
                if x == 0:
                    Parm = 'd-space'
                elif x == 1:
                    Parm = 'height'
                elif x == 2:
                    Parm = 'width'
                elif x == 3:
                    Parm = 'profile'
                # elif x ==4:
                #    Parm = 'profile_fixed'

                if orders['peak'][y][Parm] > ff.Fourier_order(PreviousParams['peak'][y][Parm]):
                    # print 'longer'
                    change_by = orders['peak'][y][Parm] * 2 + 1 - np.size(PreviousParams['peak'][y][Parm])
                    # PreviousParams['peak'][y][Parm] = np.pad(PreviousParams['peak'][y][Parm], (0,change_by),
                    # mode='constant', constant_values=0)
                    PreviousParams['peak'][y][Parm] = (PreviousParams['background'][y] + [0] * change_by)

                elif orders['peak'][y][Parm] < ff.Fourier_order(PreviousParams['peak'][y][Parm]):
                    # print 'smaller'
                    change_by = np.size(PreviousParams['peak'][y][Parm]) - ((orders['peak'][y][Parm]) * 2 + 1)
                    PreviousParams['peak'][y][Parm] = PreviousParams['peak'][y][Parm][1:-(change_by - 1)]

            if 'profile_fixed' in orders['peak'][y]:
                PreviousParams['peak'][y]['profile'] = orders['peak'][y]['profile_fixed']
            elif 'profile_fixed' in PreviousParams['peak'][y][Parm]:
                del PreviousParams['peak'][y]['profile_fixed']

        # loop for background orders/size
        for y in range(np.max([len(orders['background']), len(PreviousParams['background'])])):
            # loop over the degree of the background

            if len(PreviousParams['background']) - 1 >= y and len(orders['background']) - 1 >= y:
                # if present in both arrays make sure it is the right size.

                if orders['background'][y] > ff.Fourier_order(PreviousParams['background'][y]):
                    # print 'longer'
                    change_by = orders['background'][y] * 2 + 1 - np.size(PreviousParams['background'][y])
                    # print change_by
                    # PreviousParams['background'][y] = np.pad(PreviousParams['background'][y], (0,change_by),
                    # mode='constant', constant_values=0)
                    PreviousParams['background'][y] = (PreviousParams['background'][y] + [0] * change_by)

                elif orders['background'][y] < ff.Fourier_order(PreviousParams['background'][y]):
                    # print 'smaller'
                    change_by = np.size(PreviousParams['background'][y]) - ((orders['background'][y]) * 2 + 1)
                    PreviousParams['background'][y] = PreviousParams['background'][y][:-(change_by)]

            elif len(PreviousParams['background']) - 1 >= y > len(orders['background']) - 1:
                # if previous params has too many degrees remove the higher ones
                PreviousParams['background'][y] = []

            elif len(PreviousParams['background']) - 1 < y <= len(orders['background']) - 1:
                # if previous parameters does not have enough degrees add more.
                PreviousParams['background'].append([0] * (orders['background'][y] * 2 + 1))

        # tidy up from reductions in length of PreviousParams
        PreviousParams['background'] = [x for x in PreviousParams['background'] if x != []]
        # FIX ME: NO backg_type set here!

    if PreviousParams and not orders:
        backg = PreviousParams['background']
        backg_type = 'coeffs'

    lenbg = []
    singlelenbg = []
    for val in backg:
        lenbg.append(len(val))
        singlelenbg.append(1)
    lenbg = np.array(lenbg)
    singlelenbg = np.array(singlelenbg)

    # If there is not previous fit -- Fit data in azimuthal chunks.
    if not PreviousParams:

        if 'PeakPositionSelection' not in orders:
            # DMF changed print statement to include number peaks in inputs file
            if peeks > 1:
                p = 'peaks'
            else:
                p = 'peak'
            print('\nNo previous fits or selections. Assuming ' + str(peeks) + ' ' + p + ' and group fitting in azimuth...\n')
        else:
            print('\nUsing manual selections for initial guesses...\n')

            # FIX ME: Need to check that the number of peaks for which we have parameters is the same as the number
            # of peaks guessed at.

            tthguesses = np.array(orders['PeakPositionSelection'])
            # for future use in setting limits
            dist = np.max(tthguesses[:, 2]) - np.min(tthguesses[:, 2])
            width_change = dist / (2 ** peeks)
            peak_vals = np.unique(tthguesses[:, 2])

            # fit fourier series to two-theta/d-spacing for each peak
            # FIX ME: this fitting should be done to the d-spacing not the tth/energy. Probably need to import the
            # detector functions to make this happen.
            dfour = []
            for j in range(peeks):
                peek = j + 1
                tthguess = tthguesses[tthguesses[:, 0] == peek, 1:]
                param_str = 'peak_' + str(j)
                comp = 'd'

                temp_param = Parameters()
                temp_param = ff.initiate_params(param=temp_param, param_str=param_str, comp=comp,
                                                nterms=orders['peak'][j]['d-space'])
                # temp_param.pretty_print()
                fout = ff.Fourier_fit(azimu=tthguess[:, 0], ydata=tthguess[:, 1],
                                      param=temp_param, param_str=param_str + '_' + comp, fit_method='leastsq')
                temp_param = fout.params
                if debug:
                    temp_param.pretty_print()
                dfour.append(ff.gather_paramerrs_to_list(temp_param, param_str, comp))
                '''
                dfour.append(ff.Fourier_fit_old(tthguess[:, 0], det.Conversion(tthguess, conversion_factor),
                                                terms=orders['peak'][j]['d-space']))
                '''
        
        # get chunks according to detector type. 
        chunks, azichunks = det.bins(azimu, orders)

        # Final output list of azimuths with corresponding twotheta_0,h,w

        # wavelength=parms_dict['wavelength']
        newd0 = [[] for _ in range(peeks)]
        newd0Err = [[] for _ in range(peeks)]
        newHall = [[] for _ in range(peeks)]
        newHallErr = [[] for _ in range(peeks)]
        newWall = [[] for _ in range(peeks)]
        newWallErr = [[] for _ in range(peeks)]
        newPall = [[] for _ in range(peeks)]
        newPallErr = [[] for _ in range(peeks)]
        newBGall = [[] for _ in range(len(bg_order))]
        newBGallErr = [[] for _ in range(len(bg_order))]
        newAziChunks = []

        for j in range(len(chunks)):

            #print('\nFiting to data chunk ' + str(j + 1) + ' of ' + str(len(chunks)) + '\n')
            print('\rFitting to data chunk %s of %s ' % (str(j + 1), str(len(chunks))), end='\r')
            if debug:
                print('\n')
            # print '\nFitting azimuth chunk....\n'
            # only compute the fit and save it if there are 'enough' data
            
            min_dat = 21  #minimum data in each chunk
            n = 5 #how many data to use in constructing guesses.
            wguess_fraction = 10.
            # N.B. if min_dat/n is too large ther guesses will become very strage. 
            #FIX ME: These should be input variables (for use in extremis)
            
            # FIX ME: possibly do this better using scipy.singal.find or scipy.signal.find_peaks_cwt
             
            if np.sum(~intens.flatten()[chunks[j]].mask) >= min_dat:
                
                # get indicies of sorted two theta values excluding the masked values
                tth_ord = ma.argsort(twotheta.flatten()[chunks[j]].compressed())
                
                                             
                # background estimates
                #make bg list
                backg_guess = [[0.] for i in range(len(orders['background']))]
                
                if backg_type == 'coeffs':
                    # Fourier expand the coefficents. 
                    for j in range(len(orders['background'])):
                        backg_guess[j] = ff.Fourier_expand(np.mean(azimu.flatten()[chunks[j]]), param=orders['peak']['background'][j])
                    
                elif backg_type == 'order':
                    #first value (offset) is mean of left hand values
                    backg_guess[0][0] = np.mean(intens.flatten()[chunks[j]].compressed()[tth_ord[:n]])
                    # if there are more then calculate a gradient guess. 
                    # leave all higher order terms as 0 -- assume they are small.
                    if len(orders['background']) > 1:
                        backg_guess[1][0] = (np.mean(intens.flatten()[chunks[j]].compressed()[tth_ord[-n:]]) - np.mean(intens.flatten()[chunks[j]].compressed()[tth_ord[:n]])) / (twotheta.flatten()[chunks[j]].compressed()[tth_ord[-1]] - twotheta.flatten()[chunks[j]].compressed()[tth_ord[0]])
                        
                # elif backg_type == 'flat':
                #     backg_guess = backg[0][0]
                # Organise guesses to be refined.
                dguess = []
                hguess = []
                wguess = []
                pguess = []
                pfixed = 0
                peaks = []
                lims = []

                for k in range(peeks):
                    if 'dfour' in locals(): # if the positions have been pre guessed extract values from dfour.
                        if debug and k == 0:
                            print('dfour in locals')
                        
                        # guess returns tthguess from dfour then converts in into d-spacing.
                        tthguess = ff.Fourier_expand(np.mean(azimu.flatten()[chunks[j]]), dfour[k][0])
                        dguess   = det.Conversion(tthguess, conversion_factor, azm=np.mean(azimu.flatten()[chunks[j]]))
                        #finds the index of the closes two theta to the d-spacing input  
                        idx = ((np.abs(twotheta.flatten()[chunks[j]] - tthguess)).argmin())
                        # FIX ME: The mean of a number of the smallest values would be more stable.

                        # height is intensity of slosest pixel in d-spacing - background at that position
                        hguess = intens.flatten()[chunks[j]][idx] 
                        if len(backg_guess) > 1:
                            hguess = hguess - (backg_guess[0][0] + backg_guess[1][0] * (twotheta.flatten()[chunks[j]][idx] - twotheta.flatten()[chunks[j]].min() ))
                        elif len(backg_guess) == 1:
                            hguess = hguess - backg_guess[0][0]
                            
                        # # wguess is fractional width of the data range
                        # wguess = (det.Conversion(np.max(dspace.flatten()[chunks[j]]), conversion_factor, reverse=1, azm=np.mean(azimu.flatten()[chunks[j]])) - 
                        #           det.Conversion(np.min(dspace.flatten()[chunks[j]]), conversion_factor, reverse=1, azm=np.mean(azimu.flatten()[chunks[j]])) ) / wguess_fraction
                        # # FIX ME: This is very crude. Need a better way to do it. \

                        # pguess = 0.5  # guess half if solving for profile
                        # pfixed = 0  # set to 0 so solving unless profile_fixed exists.
                        # if 'profile_fixed' in orders['peak'][k]:
                        #     # FIX ME: profile fixed is the switch 1/0/true/false and profile is the profile parameters. Make sure this is the case in the input files
                        #     pguess = ff.Fourier_expand(np.mean(azimu.flatten()[chunks[j]]), param=orders['peak'][k]['profile_fixed']) 
                        #     pfixed = 1

                        # peaks.append({"d-space": [dguess],
                        #               "height": [hguess],
                        #               "profile": [pguess],
                        #               "width": [wguess]
                        #               })
    
                    else: # guess guesses from data slice - assuming a single peak
                        
                        # find brightest pixels
                        idx = (-intens.flatten()[chunks[j]].compressed()).argsort()[:n]
                        # height guess is nth highest intensity - backgound guess at this position
                        hguess = (intens.flatten()[chunks[j]].compressed())[idx[n-1]]
                        if len(backg_guess) > 1:
                            hguess = hguess - (backg_guess[0][0] + backg_guess[1][0] * (twotheta.flatten()[chunks[j]].compressed()[idx[n-1]] - twotheta.flatten()[chunks[j]].min() ))
                        elif len(backg_guess) == 1:
                            hguess = hguess - backg_guess[0][0]
                            
                        # d-spacing of highest nth intestity pixel
                        dguess = dspace.flatten()[chunks[j]].compressed()[idx[n - 1]]

                    # wguess is fractional width of the data range
                    wguess = (np.max(twotheta.flatten()[chunks[j]]) - np.min(twotheta.flatten()[chunks[j]])) / wguess_fraction / peeks
                    # FIX ME: This is a bit crude. Is there a better way to do it?
    
                    # if profile_fixed exists then set pguess to profile_fixed values and set pfixed to 1
                    
                    pguess = 0.5  # guess half if solving for profile
                    pfixed = 0  # set to 0 so solving unless profile_fixed exists.
                    if 'profile_fixed' in orders['peak'][k]:
                            # profile fixed is the list of coefficients if the profile is fixed. 
                            # FIX ME: profile fixed must have the same number of coefficients as required by profile.
                            pguess = ff.Fourier_expand(np.mean(azimu.flatten()[chunks[j]]) * orders['peak'][k]['symmetry'], param=orders['peak'][k]['profile_fixed']) 
                            pfixed = 1
    
                    peaks.append({"d-space": [dguess],
                                  "height": [hguess],
                                  "width": [wguess],
                                  "profile": [pguess]
                                  })
                  
                    # DMF added needs checking - required to drive fit and avoid failures
                    lims.append({"d-space": [np.min(dspace.flatten()[chunks[j]]), np.max(dspace.flatten()[chunks[j]])],
                                 "height": [0, np.max(intens.flatten()[chunks[j]])],
                                 "width": [(np.max(twotheta.flatten()[chunks[j]]) - np.min(twotheta.flatten()[chunks[j]])) / 100 / peeks,
                                           (np.max(twotheta.flatten()[chunks[j]]) - np.min(twotheta.flatten()[chunks[j]])) / 2 / peeks],
                                 "profile": [0, 1]
                                 })

                # # Guess for background, if orders and therefore no guess input, choose appropriate
                # singleBackg = [[] for i in range(len(backg))]
                # for i in range(len(backg)):
                #     singleBackg[i] = [backg[i][0]]
                # bgguess = singleBackg
                # bgguess[0][0] = 5 
                # print(bgguess)
                Guesses = {"background": backg_guess, #bgguess,
                           "peak": peaks}
                limits = {"peak": lims}
                
                 # Define parameters to pass to fit
                params = Parameters()
                # if orders bg of form [num_orders_0,num_orders_1]
                # if coeffs bg of form [[coeff1],[coeff2,coeff3,coeff4]]

                # Chunks limited to no Fourier expansion so only single value per polynomial order.
                # for b in range(len(bgguess)):
                #     params.add('bg_c' + str(b) + '_f' + str(0), bgguess[b][0])
                for b in range(len(backg_guess)):
                    params.add('bg_c' + str(b) + '_f' + str(0), backg_guess[b][0])

                for pk in range(len(Guesses['peak'])):

                    if 'dfour' in locals():
                        params.add('peak_' + str(pk) + '_d0', Guesses['peak'][pk]['d-space'][0],
                                   min=limits['peak'][pk]['d-space'][0], max=limits['peak'][pk]['d-space'][1])
                        params.add('peak_' + str(pk) + '_h0', Guesses['peak'][pk]['height'][0],
                                   min=limits['peak'][pk]['height'][0], max=limits['peak'][pk]['height'][1])
                        params.add('peak_' + str(pk) + '_w0', Guesses['peak'][pk]['width'][0],
                                   min=limits['peak'][pk]['width'][0], max=limits['peak'][pk]['width'][1])
                        if 'profile_fixed' in orders['peak'][pk]:
                            params.add('peak_' + str(pk) + '_p0', Guesses['peak'][pk]['profile'][0],
                                       min=limits['peak'][pk]['profile'][0], max=limits['peak'][pk]['profile'][1],
                                       vary=False)
                        else:
                            params.add('peak_' + str(pk) + '_p0', Guesses['peak'][pk]['profile'][0],
                                       min=limits['peak'][pk]['profile'][0], max=limits['peak'][pk]['profile'][1])


                    else:

                        # If using bounds/limits should be written as below:
                        params.add('peak_' + str(pk) + '_d0', Guesses['peak'][pk]['d-space'][0],
                                   min=limits['peak'][pk]['d-space'][0], max=limits['peak'][pk]['d-space'][1])
                        params.add('peak_' + str(pk) + '_h0', Guesses['peak'][pk]['height'][0],
                                   min=limits['peak'][pk]['height'][0], max=limits['peak'][pk]['height'][1])
                        params.add('peak_' + str(pk) + '_w0', Guesses['peak'][pk]['width'][0],
                                   min=limits['peak'][pk]['width'][0], max=limits['peak'][pk]['width'][1])
                        if 'profile_fixed' in orders['peak'][pk]:
                            params.add('peak_' + str(pk) + '_p0', Guesses['peak'][pk]['profile'][0],
                                       min=limits['peak'][pk]['profile'][0], max=limits['peak'][pk]['profile'][1],
                                       vary=False)
                        else:
                            params.add('peak_' + str(pk) + '_p0', Guesses['peak'][pk]['profile'][0],
                                       min=limits['peak'][pk]['profile'][0], max=limits['peak'][pk]['profile'][1])


                
                if debug:
                    params.pretty_print()
                    print('\n)')
                    guess = params
                out = ff.FitModel(intens.flatten()[chunks[j]], twotheta.flatten()[chunks[j]], azichunks[j],
                                  num_peaks=len(Guesses['peak']), nterms_back=len(backg_guess), Conv=conversion_factor,
                                  fixed=pfixed, method=None, weights=None, params=params)

                params = out.params
                if debug:
                    params.pretty_print()

                for i in range(peeks):
                    newd0[i].append(params['peak_' + str(i) + '_d0'].value)
                    newd0Err[i].append(params['peak_' + str(i) + '_d0'].stderr)
                    newHall[i].append(params['peak_' + str(i) + '_h0'].value)
                    newHallErr[i].append(params['peak_' + str(i) + '_h0'].stderr)
                    newWall[i].append(params['peak_' + str(i) + '_w0'].value)
                    newWallErr[i].append(params['peak_' + str(i) + '_w0'].stderr)
                    # profile should now be bound by the settings in the parameter
                    newPall[i].append(params['peak_' + str(i) + '_p0'].value)
                    newPallErr[i].append(params['peak_' + str(i) + '_p0'].stderr)
                    # may have to set profile error when initiate params so have appropriate error value if fixed
                for i in range(len(backg_guess)):
                    newBGall[i].append(params['bg_c' + str(i) + '_f0'].value)
                    newBGallErr[i].append(params['bg_c' + str(i) + '_f0'].stderr)

                newAziChunks.append(azichunks[j])

                # plot the fits.
                if debug:
                    # asdf_ord = np.argsort(twotheta.flatten()[chunks[j]])
                    # tth_plot = twotheta.flatten()[chunks[j]][asdf_ord]
                    # int_plot = intens.flatten()[chunks[j]][asdf_ord]
                    #azm_plot = azimu.flatten()[chunks[j]][asdf_ord]
                    tth_plot = twotheta.flatten()[chunks[j]]
                    int_plot = intens.flatten()[chunks[j]]
                    azm_plot = np.tile(azimu.flatten()[chunks[j]][0], (300))
                    azm_plot = ma.array(azm_plot, mask=(~np.isfinite(azm_plot))) #FIX ME: required for plotting energy dispersive data.
                    gmodel = Model(ff.PeaksModel, independent_vars=['twotheta', 'azi'])
                    tth_range = np.linspace(np.min(twotheta.flatten()[chunks[j]]), np.max(twotheta.flatten()[chunks[j]]), azm_plot.size)
                    # mod_plot = gmodel.eval(params=params, twotheta=tth_range, azi=azm_plot,
                    #                        num_peaks=len(Guesses['peak']),
                    #                        nterms_back=len(backg_guess), Conv=conversion_factor, fixed=pfixed)
                    # guess_plot = gmodel.eval(params=guess, twotheta=tth_range, azi=azm_plot,
                    #                        num_peaks=len(Guesses['peak']),
                    #                        nterms_back=len(backg_guess), Conv=conversion_factor, fixed=pfixed)
                    mod_plot = gmodel.eval(params=params, twotheta=tth_range, azi=azm_plot, Conv=conversion_factor)
                    guess_plot = gmodel.eval(params=guess, twotheta=tth_range, azi=azm_plot, Conv=conversion_factor)
                    plt.plot(tth_plot, int_plot, '.', label="data")
                    plt.plot(tth_range, guess_plot, marker='', color='green', linewidth=2, linestyle='dashed', label='guess')
                    plt.plot(tth_range, mod_plot, marker='', color='red', linewidth=2, label='fit')
                    plt.xlim(tthrange)
                    plt.legend()
                    plt.title(peak_string(orders) + '; Chunk ' + str(j+1))
                    plt.show()
                    
                    

        # Feed each d_0,h,w into fourier expansion function to get fit for fourier component
        # parameters as output.

        print('\n Doing Fourier fits...')
        # print('test0')

        dfour = []
        hfour = []
        wfour = []
        pfour = []
        master_params = Parameters()  # initiate total Parameter class to add to
        # FIX ME: parameters initiated with no limits currently 
        for j in range(peeks):

            # symmetry
            if 'symmetry' in orders['peak'][j].keys():
                symm = orders['peak'][j]['symmetry']
            else:
                symm = 1

            d0_order = orders['peak'][j]['d-space']
            h_order = orders['peak'][j]['height']
            w_order = orders['peak'][j]['width']
            p_order = orders['peak'][j]['profile']
            if 'profile_fixed' in orders['peak'][j]:
                p_fixed = orders['peak'][j]['profile_fixed']
            else:
                p_fixed = False
            # print(np.array(newAziChunks), np.array(newd0[j]), d0_order, np.array(newd0Err[j]))

            param_str = 'peak_' + str(j)  # defines peak string to start parameter name

            # Initiate parameters for each component and perform initial Fourier fit
            data_arrays = newAziChunks, newd0[j], newd0Err[j], newHall[j], newHallErr[j], newWall[j], newWallErr[j], \
                          newPall[j], newPallErr[j]
            fit_limits = [[None, None, None], [None, None, None], [None, None, None], [None, None, None]]
            # fit_limits = [[np.min(newd0),np.max(newd0),None], [0,None,None],
            #              [0,(np.max(newWall)+np.min(newWall))/2.,None], [0,1,None]] # min, max, expr in order d,
            # h, w, p

            master_params, dfour, hfour, wfour, pfour = initial_component_fits(master_params, param_str, data_arrays,
                                                                               [dfour, hfour, wfour, pfour],
                                                                               [d0_order, h_order, w_order, p_order],
                                                                               symm=symm, limits=fit_limits, pfixed=p_fixed)
            if debug:
                print('Parameters after initial Fourier fits')
                master_params.pretty_print()

        newBGall = np.array(newBGall)
        newBGallErr = np.array(newBGallErr)
        bgfour = []


        # Initiate background parameters
        comp = 'bg'
        for b in range(len(backg)):
            for f_b in range(len(backg[b])):
                master_params.add('bg_c' + str(b) + '_f' + str(f_b), backg[b][f_b])  # need limits

        # Perform initial background Fourier fit
        for i in range(len(lenbg)):
            param_str = 'bg_c' + str(i)
            master_params = ff.unvary_params(master_params, param_str, '')
            master_params = ff.vary_params(master_params, param_str, '')
            fout = ff.Fourier_fit(azimu=np.array(newAziChunks), ydata=np.array(newBGall[i]), param=master_params,
                                  errs=np.array(newBGallErr[i]), param_str=param_str, fit_method='leastsq')
            master_params = fout.params
            bgfour.append(ff.gather_paramerrs_to_list(master_params, param_str, 'f'))

        # plot output of fourier fits....
        if debug:
            y_lims = np.array([np.min(newAziChunks), np.max(newAziChunks)])
            y_lims = np.around(y_lims / 180) * 180
            AziPlot = range(np.int(y_lims[0]), np.int(y_lims[1]), 2)
            gmodel = Model(ff.Fourier_expand, independent_vars=['azimu'])

            fig = plt.figure()
            ax1 = fig.add_subplot(5, 1, 1)
            ax1.set_title('D-spacing')
            for j in range(peeks):
                ax1.scatter(newAziChunks, newd0[j], s=10)
                gmod_plot = gmodel.eval(params=master_params, azimu=np.array(AziPlot), param=dfour[j][0])
                ax1.plot(AziPlot, gmod_plot)

            # plt.subplot(512)
            ax2 = fig.add_subplot(5, 1, 2)
            ax2.set_title('height')
            for j in range(peeks):
                if 'symmetry' in orders['peak'][j].keys():
                    symm = orders['peak'][j]['symmetry']
                else:
                    symm = 1
                ax2.scatter(newAziChunks, newHall[j], s=10)
                gmod_plot = gmodel.eval(params=master_params, azimu=np.array(AziPlot) * symm, param=hfour[j][0])
                ax2.plot(AziPlot, gmod_plot)

            ax3 = fig.add_subplot(5, 1, 3)
            # plt.subplot(513)
            ax3.set_title('width')
            for j in range(peeks):
                if 'symmetry' in orders['peak'][j].keys():
                    symm = orders['peak'][j]['symmetry']
                else:
                    symm = 1
                ax3.scatter(newAziChunks, newWall[j], s=10)
                gmod_plot = gmodel.eval(params=master_params, azimu=np.array(AziPlot) * symm, param=wfour[j][0])
                ax3.plot(AziPlot, gmod_plot)

            ax4 = fig.add_subplot(5, 1, 4)
            # plt.subplot(514)
            ax4.set_title('Profile')
            for j in range(peeks):
                if 'symmetry' in orders['peak'][j].keys():
                    symm = orders['peak'][j]['symmetry']
                else:
                    symm = 1
                ax4.scatter(newAziChunks, newPall[j], s=10)
                gmod_plot = gmodel.eval(params=master_params, azimu=np.array(AziPlot) * symm, param=pfour[j][0])
                ax4.plot(AziPlot, gmod_plot)

            for k in range(len(lenbg)):
                x_plt = len(lenbg)
                y_plt = len(lenbg) * 4 + k + 1
                ax5 = fig.add_subplot(5, x_plt, y_plt)
                ax5.set_title('Background')
                ax5.scatter(newAziChunks, newBGall[k], s=10)
                gmod_plot = gmodel.eval(params=master_params, azimu=np.array(AziPlot), param=bgfour[k][0])
                ax5.plot(AziPlot, gmod_plot)

            fig.suptitle(peak_string(orders) + '; Fits to Chunks')
            plt.show()
            plt.close()

        # put fits into NewParams data structure.
        NewParams = ff.create_newparams(peeks, dfour, hfour, wfour, pfour, bgfour, orders['peak'])

        # Iterate over each fourier series in turn.

        if refine:
            print('\nRe-fitting for d, h, w, bg separately...\n')
            # iterate through the variables.
            for j in range(iterations):

                dfour = []
                hfour = []
                wfour = []
                pfour = []
                # refine parameters for each peak.
                for k in range(peeks):

                    param_str = 'peak_' + str(k)
                    if 'profile_fixed' in orders['peak'][k]:
                        p_fixed = True
                    else:
                        p_fixed = False
                    if 'symmetry' in orders['peak'][k]:
                        symm = orders['peak'][k]['symmetry']
                    else:
                        symm = 1
                    master_params, dfour, hfour, wfour, pfour = update_component_fits(master_params, intens, azimu,
                                                                                      twotheta, param_str,
                                                                                      [dfour, hfour, wfour, pfour],
                                                                                      conversion_factor, backg_guess,
                                                                                      Guesses, p_fixed)

                # refine background
                param_str = 'bg_c'
                # print(param_str, lenbg, i)
                master_params = ff.unvary_params(master_params, param_str, '')
                master_params = ff.vary_params(master_params, param_str, '')
                out = ff.FitModel(intens.flatten(), twotheta.flatten(), azimu.flatten(), num_peaks=len(Guesses['peak']),
                                  nterms_back=len(backg_guess), Conv=conversion_factor, fixed=pfixed, method=None,
                                  weights=None, params=master_params)
                master_params = out.params
                # Put fits into data structure.
                NewParams = ff.create_newparams(peeks, dfour, hfour, wfour, pfour, bgfour, orders['peak'])
        
                if debug:
                    print('Parameters after refining Fourier fits ' + str(j) + 'times')
                    master_params.pretty_print()


    else:
        # FIX ME: This should load the saved lmfit Parameter class object. Need to initiate usage (save and load).
        # For now have propagated use of NewParams so re-instantiate the master_params object below, which is clunky.
        print('Using previously fitted parameters and propagating fit')
        NewParams = PreviousParams

        # FIX ME: should this be some def or subfunction?
        master_params = Parameters()  # initiate total Parameter class to add to
        # FIX ME: parameters initiated with no limits currently 

        for j in range(peeks):

            param_str = 'peak_' + str(j)  # defines peak string to start parameter name

            # initiate symmetry
            comp = 's'
            if 'symmetry' in orders['peak'][j].keys():
                symm = orders['peak'][j]['symmetry']
            else:
                symm = 1
            master_params = ff.initiate_params(master_params, param_str, comp, 0, max=10, min=1, value=np.array(symm))

            # initiate d
            comp = 'd'
            dfour = PreviousParams['peak'][j]['d-space']
            master_params = ff.initiate_params(master_params, param_str, comp, orders['peak'][j]['d-space'],
                                               value=np.array(dfour))

            # initiate h
            comp = 'h'
            hfour = PreviousParams['peak'][j]['height']
            master_params = ff.initiate_params(master_params, param_str, comp, orders['peak'][j]['height'], value=hfour)

            # initiate w
            comp = 'w'
            wfour = PreviousParams['peak'][j]['width']
            master_params = ff.initiate_params(master_params, param_str, comp, orders['peak'][j]['width'], value=wfour)

            # initiate p
            comp = 'p'
            if orders and 'profile_fixed' in orders['peak'][j]:
                pfour = orders['peak'][j]['profile_fixed']
                pfixed = 1
            else:
                pfour = PreviousParams['peak'][j]['profile']
                pfixed = 0
            master_params = ff.initiate_params(master_params, param_str, comp, orders['peak'][j]['profile'], value=pfour)

        # Initiate background parameters
        comp = 'bg'
        for b in range(len(backg)):
            # print(backg[b])
            for f_b in range(len(backg[b])):
                # print(b, f_b)
                master_params.add('bg_c' + str(b) + '_f' + str(f_b),
                                  PreviousParams['background'][b][f_b])  # need limits

        # master_params.pretty_print()

        # Guess for background, if orders and therefore no guess input, choose appropriate
        # copied from above to allow the code to run. FIX ME: doesn't need all this code.
        singleBackg = [[] for i in range(len(backg))]
        for i in range(len(backg)):
            singleBackg[i] = [backg[i][0]]
        backg_guess = singleBackg
        #bgguess[0][0] = 5

        # def previous_params_to_params(PreviousParams):

        # FIX ME: need to confirm the number of parameters matches the orders of the fits.

    # FIX ME: Need to sort out use of lmfit parameters and NewParams, below will not work currently for loaded
    # parameters as master_params not created from NewParams yet. If possible use the json import/export within lmfit.

    # Refit through full equation with all data for d,h,w,bg independently
    print('\nFinal fit solving for all parms...\n')

    num_peaks = len(orders['peak'])

    # set all parameters to vary
    for p in master_params:
        master_params[p].set(vary=True)
    for x in range(num_peaks):
        # unvary symmetry!! This should never vary.
        new_str = 'peak_' + str(x) + '_s'
        str_keys = [key for key, val in master_params.items() if new_str in key]
        for pstr in str_keys:
            master_params[pstr].set(vary=False)

        # unvary p for appropriate peaks
        if 'profile_fixed' in orders['peak'][x]:
            new_str = 'peak_' + str(x) + '_p'
            str_keys = [key for key, val in master_params.items() if new_str in key]
            for pstr in str_keys:
                master_params[pstr].set(vary=False)
    
    # FIX ME: need to sort parameters, bgguess doesn't exist if load previous fit data
    out = ff.FitModel(intens.flatten(), twotheta.flatten(), azimu.flatten(), num_peaks=num_peaks,
                      nterms_back=len(backg_guess), Conv=conversion_factor, fixed=pfixed, method=None,
                      weights=None, params=master_params)
    master_params = out.params
    
    print('\nFinal Coefficients\n')
    # print(out.fit_report(min_correl=1))
    print(out.fit_report())
    # print('Final Coefficients\n', NewParams)
    
    # Write master_params to NewParams dict object
    NewParams = ff.params_to_newparams(master_params, num_peaks=num_peaks, num_bg=len(backg_guess),
                                       order_peak=orders['peak'])

    # Stats now directly from lmfit info.
    # number elements
    n_points = out.ndata
    print('number data', n_points)

    # degrees of freedom (number of parameters solved for).
    deg_freedom = out.nfree
    print('degrees of freedom', deg_freedom)

    # ChiSq from fit
    ChiSq = out.chisqr
    print('ChiSquared', ChiSq)

    '''
    # Calculate SSD for fit.
    # inp, pen = ff.PeaksModel(twotheta.flatten(), azimu.flatten(), NewParams, Conv=conversion_factor)
    SSD = np.sum((intens.flatten() - inp.flatten()) ** 2)
    print('SSD', SSD)

    # https://autarkaw.org/2008/07/05/finding-the-optimum-polynomial-order-to-use-for-regression/
    # We choose the degree of polynomial for which the variance as computed by
    # Sr(m)/(n-m-1)
    # is a minimum or when there is no significant decrease in its value as the degree of polynomial is increased.
    # In the above formula, Sr(m) = sum of the square of the residuals for the mth order polynomial
    # n= number of data points
    # m=order of polynomial (so m+1 is the number of constants of the model)

    # number elements
    n_points = intens.flatten().size
    print('number data', n_points)

    # degrees of freedom (number of parameters solved for).
    deg_freedom = sum(ff.flatten(Chng))
    print('degrees of freedom', deg_freedom)

    SSD_var = SSD / (n_points + deg_freedom - 1)
    print('variance', SSD_var)

    # ChiSq is defined in the same way as GSAS. See GSAS manual page 165/166
    ChiSq = SSD / (n_points - deg_freedom)
    print('ChiSquared', ChiSq)
    '''

    # Plot results to check
    view = 0
    if SaveFit == 1 or view == 1 or debug:
        print('\nPlotting results for fit...\n')

        gmodel = Model(ff.PeaksModel, independent_vars=['twotheta', 'azi'])
        # fullfit_intens = gmodel.eval(params=master_params, twotheta=twotheta.flatten(), azi=azimu.flatten(),
        #                              num_peaks=num_peaks, nterms_back=len(backg_guess), Conv=conversion_factor,
        #                              fixed=pfixed)
        fullfit_intens = gmodel.eval(params=master_params, twotheta=twotheta.flatten(), azi=azimu.flatten(),
                                     Conv=conversion_factor)
        # fullfit_intens = inp
        # Set the maximum to the n+1th value in the array
        # FIX ME: the max find is not necessarily efficient. A quicker way should be found if it exists.
        max_pos = 5

        ValMax1 = -np.sort(-intens.flatten())[max_pos]  # np.max(intens.flatten())
        ValMax2 = -np.sort(-fullfit_intens.flatten())[max_pos]  # np.max(fullfit_intens.flatten())
        ValMax3 = -np.sort(-(intens.flatten() - fullfit_intens.flatten()))[max_pos]
        # np.max((intens.flatten()-fullfit_intens.flatten()))
        ValMax = np.max([ValMax1, ValMax2])
        ValMin1 = np.min(intens.flatten())
        ValMin2 = np.min(fullfit_intens.flatten())
        ValMin3 = np.min((intens.flatten() - fullfit_intens.flatten()))
        ValMin = np.min([ValMin1, ValMin2])

        y_lims = np.array([np.min(azimu.flatten()), np.max(azimu.flatten())])
        y_lims = np.around(y_lims / 180) * 180

        if conversion_factor['DispersionType'] == 'EnergyDispersive':
            xlabel_str = 'Energy (keV)'
        elif conversion_factor['DispersionType'] == 'AngleDispersive':
            xlabel_str = 'Two theta (deg)'
        else:
            xlabel_str = 'Dispersive units'

        dot_size = 16

        fig = plt.figure()
        axf = fig.add_subplot(1, 3, 1)
        axO1 = plt.subplot(131)
        # plt.scatter(twotheta, azimu, s=1, c=np.log(intens), edgecolors='none', cmap=plt.cm.jet)
        plt.scatter(twotheta.flatten(), azimu.flatten(), s=dot_size, c=(intens.flatten()), edgecolors='none',
                    cmap=plt.cm.magma_r, vmin=ValMin, vmax=ValMax)
        axO1.set_title('Data')
        axO1.set_xlabel(xlabel_str)
        axO1.set_ylabel('Azimuth (deg)')
        axO1.set_xlim([np.min(twotheta.flatten()), np.max(twotheta.flatten())])
        axO1.set_ylim(y_lims)
        locs, labels = plt.xticks()
        plt.setp(labels, rotation=90)
        plt.colorbar()
        axO2 = plt.subplot(132)
        # plt.scatter(twotheta, azimu, s=1, c=np.log(fullfit_intens), edgecolors='none', cmap=plt.cm.jet)
        plt.scatter(twotheta.flatten(), azimu.flatten(), s=dot_size, c=(fullfit_intens.flatten()), edgecolors='none',
                    cmap=plt.cm.magma_r, vmin=ValMin, vmax=ValMax)
        plt.colorbar()
        axO2.set_title('Model')
        axO2.set_xlabel(xlabel_str)
        axO2.set_ylabel('Azimuth (deg)')
        axO2.set_xlim([np.min(twotheta.flatten()), np.max(twotheta.flatten())])
        axO2.set_ylim(y_lims)
        locs, labels = plt.xticks()
        plt.setp(labels, rotation=90)
        axO3 = plt.subplot(133)
        # plt.scatter(twotheta, azimu, s=1, c=np.log(intens-fullfit_intens), edgecolors='none', cmap=plt.cm.jet)
        plt.scatter(twotheta.flatten(), azimu.flatten(), s=dot_size, c=(intens.flatten() - fullfit_intens.flatten()),
                    edgecolors='none', cmap=plt.cm.coolwarm, vmin=ValMin3, vmax=ValMax3)
        axO3.set_title('Residuals (data - model)')
        axO3.set_xlabel(xlabel_str)
        axO3.set_ylabel('Azimuth (deg)')
        axO3.set_xlim([np.min(twotheta.flatten()), np.max(twotheta.flatten())])
        axO3.set_ylim(y_lims)
        locs, labels = plt.xticks()
        plt.setp(labels, rotation=90)
        plt.colorbar()
        plt.suptitle(peak_string(orders) + '; final fit')
        plt.tight_layout()

        #        # make a second figure with the Q-Q plot of the data and fit.
        #        #the second subplot is a cdf of the residuals with a gaussian added to it.
        #        fig2 = plt.figure()
        #        axQ = plt.subplots()
        #        plt.scatter(intens.flatten(), fullfit_intens.flatten(), s=dot_size)
        #        #axQ.set_xlabel('Observations')
        #        #axQ.set_ylabel('Model')
        #        plt.grid(True)
        #        plt.tight_layout()

        # save figures without overwriting old names
        if SaveFit:
            if not fnam == None:
                filename = os.path.splitext(os.path.basename(fnam))[0] + '_'
            else:
                filename = 'Fit2Peak_'
            filename = filename + peak_string(orders, fname=True)

            i = 0
            if os.path.exists('{}.png'.format(filename)):
                i+=1
            while os.path.exists('{}_{:d}.png'.format(filename, i)):
                i += 1
            if i == 0:
                plt.savefig('{}.png'.format(filename))
            else:
                plt.savefig('{}_{:d}.png'.format(filename, i))

        if view == 1 or debug:
            plt.show()

        plt.close()
        print('Done with Figure')

    #save lmfit structure
    if SaveFit:
        if not fnam == None:
            filename = os.path.splitext(os.path.basename(fnam))[0] + '_'
        else:
            filename = 'Fit2Peak_'
        filename = filename + peak_string(orders, fname=True)
                
        try:
            save_modelresult(out, filename+'.sav')
        except:
            print('Cannot save lmfit object')
            if os.path.isfile(filename+'.sav'):
                os.remove(filename+'.sav')
            else:    
                print("File doesn't exists!")
        
    tth_min = twotheta.min()
    tth_max = twotheta.max()
    d_min = dspace.min()
    d_max = dspace.max()
    extent = [[tth_min, tth_max], [d_min, d_max]]  # FIX ME: This is taking the maximum and the minimum of the data
    # not the 'range' itself.

    NewParams.update({'range': extent})

    FitStats = {"degree-of-freedom": deg_freedom,
                "n-points": n_points,
                "ChiSq": ChiSq
                }

    NewParams.update({'FitProperties': FitStats})

    # FIX ME: Need to put master_params into NewParams to return!

    return [ NewParams, out]



#!/usr/bin/env python

__all__ = ['create_newparams', 'params_to_newparams', 'ParamFitArr', 'PseudoVoigtPeak', 'gather_paramerrs_to_list',
           'gather_params_to_list', 'gather_params_from_dict', 'initiate_params', 'PeaksModel', 'vary_params',
           'unvary_params', 'FitModel', 'Fourier_expand', 'Fourier_fit', 'ChangeParams',
           'Fourier_backgrnd']

import numpy as np
import numpy.ma as ma
import sys
from scipy.optimize import curve_fit
from lmfit import minimize, report_fit, Model

np.set_printoptions(threshold=sys.maxsize)

# Fitting functions required for fitting peaks to the data. 
# Called by FPF_XRD_FitSubpattern.


# Known limitations/bugs
# FIX ME: need hard limits to gauss - lorentz peak profile ratios


def create_newparams(num_peaks, dfour, hfour, wfour, pfour, bgfour, order_peak):
    """
    Create the NewParams dictionary from results
    :param num_peaks: number of peaks int
    :param dfour: list of coefficients per peak float
    :param hfour: list of coefficients per peak float
    :param wfour: list of coefficients per peak float
    :param pfour: list of coefficients per peak float
    :param bgfour:
    :param order_peak:
    :return: NewParams dictionary
    """
    peaks = []
    for j in range(num_peaks):
        if 'symmetry' in order_peak[j]:  # orders['peak'][j]:  # FIX ME: the symmetry part of this is a horrible way
            # of doing it.
            peaks.append({"d-space": dfour[j][0],
                          "height": hfour[j][0],
                          "width": wfour[j][0],
                          "profile": pfour[j][0],
                          "symmetry": order_peak[j]['symmetry']})
                            # "symmetry": orders['peak'][j]['symmetry']})
        else:
            peaks.append({"d-space": dfour[j][0],
                          "height": hfour[j][0],
                          "width": wfour[j][0],
                          "profile": pfour[j][0]})
    NewParams = {"background": bgfour, "peak": peaks}

    return NewParams


def params_to_newparams(params, num_peaks, num_bg, order_peak):
    """
    Gather Parameters class to NewParams dictionary format
    :param params: lmfit Parameters class object
    :param num_peaks: number of peaks
    :param num_bg: number of background terms
    :return: NewParams dictionary object
    """

    peaks = []
    for i in range(num_peaks):
        new_str = 'peak_'+str(i)+'_d'
        dspace = gather_paramerrs_to_list(params, new_str)
        new_str = 'peak_'+str(i)+'_h'
        hspace = gather_paramerrs_to_list(params, new_str)
        new_str = 'peak_'+str(i)+'_w'
        wspace = gather_paramerrs_to_list(params, new_str)
        new_str = 'peak_'+str(i)+'_p'
        pspace = gather_paramerrs_to_list(params, new_str)

        if 'symmetry' in order_peak[i]:  # orders['peak'][j]:  # FIX ME: the symmetry part of this is a horrible
            # way of doing it.
            peaks.append({"d-space":     dspace[0],
                          "d-space_err": dspace[1],
                          "height":      hspace[0],
                          "height_err":  hspace[1],
                          "width":       wspace[0],
                          "width_err":   wspace[1],
                          "profile":     pspace[0],
                          "profile_err": pspace[1],
                          "symmetry":    order_peak[i]['symmetry']})
        else:
            peaks.append({"d-space": dspace[0],
                          "d-space_err": dspace[1],
                          "height":      hspace[0],
                          "height_err":  hspace[1],
                          "width":       wspace[0],
                          "width_err":   wspace[1],
                          "profile":     pspace[0],
                          "profile_err": pspace[1]})
    # Get background parameters
    bgspace = []
    bgspace_err = []
    for b in range(num_bg):
        new_str = 'bg_c'+str(b)+'_f'
        bgspc = gather_paramerrs_to_list(params, new_str)
        bgspace.append(bgspc[0])
        bgspace_err.append(bgspc[1])
        # print('background', ff.gather_params_to_list(params, new_str))
        # bgspace.append(ff.gather_params_to_list(params, new_str))

    # NewParams = {"background": bgspace, "peak": peaks}
    NewParams = {"background": bgspace, "background_err": bgspace_err, "peak": peaks}

    return NewParams


def flatten(li):
    return sum(([x] if not isinstance(x, list) else flatten(x)
                for x in li), [])


def ParamFitArr(numP, LenBG, val=None, p=None, b=None, orda=None):
    # Makes array which represents all the fourier series in the fit.
    # Arr{0} is a list of 4 element lists - one 4 element list for each peak.
    # Arr[0] is the background representative.

    if val:
        v = val
    else:
        v = 0

    if not p:
        p = [v, v, v, v]

    Arr = [[], []]
    for y in range(numP):
        Arr[0].append(p[:])
    for y in range(LenBG):
        Arr[1].extend([v][:])
    return Arr


# Gausian shape
def GaussianPeak(twotheta, tth0, W_all, H_all):
    W_all = W_all / np.sqrt(np.log(4))
    GaussPeak = H_all * np.exp((-(twotheta - tth0) ** 2) / (2 * W_all ** 2))
    return GaussPeak


# Lorentz shape.
def LorentzianPeak(twotheta, tth0, W_all, H_all):
    LPeak = H_all * W_all ** 2 / ((twotheta - tth0) ** 2 + W_all ** 2)
    return LPeak


# Pseudo-Voigt.
def PseudoVoigtPeak(twotheta, tth0, W_all, H_all, LGratio):
    # FIX ME: how should the ratio be fixed?
    # The ratio is between 1 and 0. it has to be limited.
    # But need options for fixed ratio and not fitting them.
    # How do I do this?
    #
    # I have defined a new variable called profile_fixed which fixes the value if present
    # if np.any(LGratio<0) or np.any(LGratio>1):
    # print 'The ratio is out of range'
    # stop

    PVpeak = LGratio * GaussianPeak(twotheta, tth0, W_all, H_all) + (1 - LGratio) * LorentzianPeak(twotheta, tth0,
                                                                                                   W_all, H_all)
    return PVpeak


def gather_paramerrs_to_list(param, param_str, comp=None, nterms=None):  # fix me: Doesn't run without Comp=None.
    # Maybe it was never validated before SAH ran it.
    """
    Make a nested list of parameters and errors from a lmfit Parameters class
    :param param: dict with multiple coefficients per component
    :param param_str: base string to select parameters
    :param comp: component to add to base string to select parameters
    :return: nested list of [parameters, errors] in alphanumerical order
    """

    if comp:
        new_str = param_str + '_' + comp
    else:
        new_str = param_str
    param_list = []
    err_list = []
    total_params = []
    str_keys = [key for key, val in param.items() if new_str in key]
    # print('yes', new_str, str_keys)
    for i in range(len(str_keys)):
        param_list.append(param[new_str + str(i)].value)
        err_list.append(param[new_str + str(i)].stderr)
    total_params.append(param_list)
    total_params.append(err_list)
    return total_params


def gather_params_to_list(param, param_str, comp=None):
    """
    Make a list of parameters from a lmfit Parameters class
    :param param: dict with multiple coefficients per component
    :param param_str: base string to select parameters
    :param comp: component to add to base string to select parameters
    :return: list of parameters in alphanumerical order
    """

    if comp:
        new_str = param_str + '_' + comp
    else:
        new_str = param_str
    param_list = []
    str_keys = [key for key, val in param.items() if new_str in key]
    # print('yes', new_str, str_keys)
    for i in range(len(str_keys)):
        param_list.append(param[new_str + str(i)].value)
    return param_list


def gather_params_from_dict(param, param_str, comp):
    """
    Make a list of parameters from a dictionary
    :param param: dict with multiple coefficients per component
    :param param_str: base string to select parameters
    :param comp: component to add to base string to select parameters
    :return: list of parameters in alphanumerical order
    """

    if comp:
        new_str = param_str + '_' + comp
    else:
        new_str = param_str
    param_list = []
    str_keys = [key for key, val in param.items() if new_str in key]
    for i in range(len(str_keys)):
        param_list.append(param[new_str + str(i)])
    return param_list


def initiate_params(param, param_str, comp, nterms=None, max=None, min=None, expr=None, value=1.):
    """
    Create all required coefficients for no. terms
    :param param: lmfit Parameter class (can be empty)
    :param param_str: base string to create lmfit parameter
    :param comp: string to add to base string to create lmfit parameter
    :param nterms: number of terms required
    :return:
    """
    
    value = np.array(value)  # force input to be array so that can be iterated over

    if nterms is None:
        nterms = int((len(value) - 1)/2)

    for t in range(2 * nterms + 1):
        if value.size == 1:
            ind = 0
        else:
            ind = t
        param.add(param_str + '_' + comp + str(t), value.item(ind), max=max, min=min, expr=expr)
        # param.add(param_str + '_' + comp + str(t), value, max=max, min=min, expr=expr)
    return param


def unvary_params(param, param_str, comp):
    """
    Set all but selected parameters to vary=False
    :param param: lmfit Parameter class
    :param param_str: base string to select with
    :param comp: string to add to base string to select with
    :return: updated lmfit Parameter class
    """
    if comp:
        new_str = param_str + '_' + comp
    else:
        new_str = param_str
    str_keys = [key for key, val in param.items() if new_str in key]
    # print(str_keys)
    for p in param:
        # print(p)
        if p not in str_keys:
            # print('yes')
            param[p].set(vary=False)
    return param


def vary_params(param, param_str, comp):
    """
    Set selected parameters to vary=True
    :param param: lmfit Parameter class
    :param param_str: base string to select with
    :param comp: string to add to base string to select with
    :return: updated lmfit Parameter class
    """
    if comp:
        new_str = param_str + '_' + comp
    else:
        new_str = param_str
    str_keys = [key for key, val in param.items() if new_str in key]
    # print(str_keys)
    for p in str_keys:
        # print('yes')
        param[p].set(vary=True)
    return param


#def PeaksModel(twotheta, azi, num_peaks=None, nterms_back=None, Conv=None, **params):
def PeaksModel(twotheta, azi, Conv=None, **params):
    """Full model of intensities at twotheta and azi given input parameters
    :param twotheta: arr values float
    :param azi: arr values float
    :param num_peaks: total number of peaks int
    :param nterms_back: total number of polynomial expansion components for the background int
    :param Conv: inputs for the conversion call dict
    :param PenCalc: penalise background or not int
    :param params: lmfit parameter class dict
    :return lmfit model fit result
    """

    # N.B. params now doesn't persist as a parameter class, merely a dictionary, so e.g. call key/value pairs as
    # normal not with '.value'

    # recreate backg array to pass to Fourier_backgrnd
    back_keys = [key for key, val in params.items() if 'bg_c' in key]
    nterm_fouriers = []
    
    i=0
    backg=[]
    while sum(nterm_fouriers) < len(back_keys):
        
        f = sum('bg_c' + str(i) in L for L in back_keys)
        nterm_fouriers.append(f)
        
        fbg = []
        for k in range(f):
            fbg.append(params['bg_c' + str(i) + '_f' + str(k)])
        backg.append(np.array(fbg))   
        i = i+1
    
    # for i in range(int(nterms_back)):
    #     f = sum('bg_c' + str(i) in L for L in back_keys)
    #     nterm_fouriers.append(f)
    # backg = []
    # for j in range(int(nterms_back)):
    #     fbg = []
    #     for k in range(nterm_fouriers[j]):
    #         fbg.append(params['bg_c' + str(j) + '_f' + str(k)])
    #     backg.append(np.array(fbg))
    # print(backg)
    
    # for future logging
    # print('recreate backg to pass to fourier_backg\n', 'back_keys is ', back_keys)
    # print('backg is ', backg)

    I = Fourier_backgrnd((azi, twotheta), backg)

    
    peak_keys = [key for key, val in params.items() if 'peak' in key]
    num_peaks = 0
    nterm_peaks = 0
    while nterm_peaks < len(peak_keys):
        f = sum('peak_' + str(num_peaks) in L for L in peak_keys)
        nterm_peaks = nterm_peaks + f
        num_peaks = num_peaks+1

    # loop over the number of peaks
    Ipeak = []
    for a in range(int(num_peaks)):

        if 'peak_' + str(a) + '_s0' in params:
            symm = params['peak_' + str(a) + '_s0']
        else:
            symm = 1

        # May need to fix for mutliple Fourier coefficients per component
        param_str = 'peak_' + str(a)
        comp = 'd'
        parms = gather_params_from_dict(params, param_str, comp)
        # print(parms)
        Dall = Fourier_expand(azi, param=parms)
        comp = 'h'
        parms = gather_params_from_dict(params, param_str, comp)
        Hall = Fourier_expand(azi * symm, param=parms)
        comp = 'w'
        parms = gather_params_from_dict(params, param_str, comp)
        Wall = Fourier_expand(azi * symm, param=parms)
        comp = 'p'
        parms = gather_params_from_dict(params, param_str, comp)
        Pall = Fourier_expand(azi * symm, param=parms)

        # conversion
        TTHall = CentroidConversion(Conv, Dall, azi)
        Ipeak.append(PseudoVoigtPeak(twotheta, TTHall, Wall, Hall, Pall))
        I = I + Ipeak[a]

    return I


def ChangeParams(constants, *fitparam):
    print('In changeparam', len(constants))
    twotheta = constants[0]
    azi = constants[1]
    ChangeArray = constants[2]
    Shapes = constants[3]
    Conv = constants[4]
    symm = constants[5]
    fixed = constants[6]

    print('changeparams:')
    print(twotheta, azi)
    print(ChangeArray)
    print(Shapes)
    print(Conv)
    print(symm)
    print(fixed)
    print(fitparam)
    Shapes = GuessApply(ChangeArray, Shapes, fitparam)
    # print(stop)
    print('In changes', Shapes)
    I, pen = PeaksModel_old(twotheta, azi, Shapes, Conv, symm, PenCalc=1)

    return I  # *pen


def FitModel(Intfit, twotheta, azimu, params, num_peaks, nterms_back, Conv=None, fixed=None,
             method='leastsq', weights=None):
    # ChangeArray, Shapes, Conv=None, symm=None, fixed=None, method=None,
    # weights=None, bounds=None):
    """Initiate model of intensities at twotheta and azi given input parameters and fit
    :param Intfit: intensity values to fit arr
    :param twotheta: twotheta values arr
    :param azimu: azimu values arr
    :param params: lmfit Parameter class dict of parameters to fit
    :param num_peaks: total number of peaks in model int
    :param nterms_back: total number of polynomial component terms in background int
    :param Conv: inputs for the conversion call dict
    :param fixed: unsure
    :param method: lmfit choice of fitting method e.g. least squares
    :param weights: errors on intensity values arr of size Intfit
    :return: lmfit model result
    """

    # bounds passed but not used

    # p0array = GuessGet(ChangeArray, Shapes)
    # print('here', Intfit, twotheta, azimu, ChangeArray, Shapes)
    # params.pretty_print()
    gmodel = Model(PeaksModel, independent_vars=['twotheta', 'azi'])
    # print('parameter names: {}'.format(gmodel.param_names))
    # print('independent variables: {}'.format(gmodel.independent_vars))

    # minner = Minimizer(fcn2min, params, fcn_args=(x, data))
    # result = minner.minimize()
    # out = gmodel.fit(Intfit, params, twotheta=twotheta, azi=azimu, num_peaks=num_peaks, nterms_back=nterms_back,
    #                  Conv=Conv, fixed=fixed, nan_policy='propagate')
    out = gmodel.fit(Intfit, params, twotheta=twotheta, azi=azimu, Conv=Conv, nan_policy='propagate')
    # print(out.fit_report())
    # print(out.params)
    # out.params.pretty_print()
    # result = popt.minimize()
    return out




def CentroidConversion(Conv, args_in, azi):
    # Conv['DispersionType'] is the conversion type
    # Conv[...] are the values required for the conversion
    # FIX ME: these functions should be a sub function of the detector types. but need to work out how to do it.

    if Conv['DispersionType'] is None or Conv == 0:
        args_out = args_in

    elif Conv['DispersionType'] == 'AngleDispersive':
        # args_in are d-spacing
        # args_outis two thetas
        args_out = 2 * np.degrees(
            np.arcsin(Conv['conversion_constant'] / 2 / args_in))  # FIX ME: check this is correct!!!

    elif Conv['DispersionType'] == 'EnergyDispersive':

        args_out = []
        # for x in range(len(azi)):
        for x in range(azi.size):
            # determine which detector is being used
            # a=np.asscalar(np.where(Conv['azimuths'] == azi[x])[0])

            if azi.size == 1:
                a = np.array(np.where(Conv['azimuths'] == azi)[0][0])
                # a= np.asscalar(np.where(Conv['azimuths'] == azi)[0][0])
                args_out.append(
                    12.398 / (2 * args_in * np.sin(np.radians(Conv['calibs'].mcas[a].calibration.two_theta / 2))))
            else:
                #                print type(azi)
                #                print azi.mask[x]
                if not azi.mask[x]:
                    a = (np.where(Conv['azimuths'] == azi[x])[0][0])
                    args_out.append(12.398 / (
                            2 * args_in[x] * np.sin(np.radians(Conv['calibs'].mcas[a].calibration.two_theta / 2))))
                else:
                    args_out.append(0)
        if isinstance(azi, np.ma.MaskedArray):
            args_out = ma.array(args_out, mask=azi.mask)
    else:
        raise ValueError('Unrecognised conversion type')

    return args_out


def Fourier_order(params):
    # Given list of Fourier coefficients return order (n) of the Fourier series.

    if isinstance(params, (list,)):
        order = int((len(params) - 1) / 2)
    elif isinstance(params, (float,)):
        order = int((np.size(params) - 1) / 2)
    elif isinstance(params, (int,)):
        order = 0
    else:
        print(params)
        print(type(params))
        raise ValueError('Parameter list is not list or float.')

    return order


# fourier expansion function
def Fourier_expand(azimu, param=None, comp_str=None, **params):
    """Calculate Fourier expansion given input coefficients
    :param azimu: arr data array float
    :param param: list of Fourier coefficients float
    :param comp_str: str to determine which coefficients to use from params
    :param params: lmfit dict of coefficients as parameters
    :return:
    """

    if param:
        # FIX ME: Need to check the fourier is a licit length
        if not isinstance(param, np.float64):
            param = np.array(param)
            if len(param.shape) > 1:
                if np.any(np.array(param.shape) > 1):
                    param = np.squeeze(param)
                elif np.all(np.array(param.shape) == 1):
                    param = np.squeeze(param)
                    param = np.array([param], float)
        # param = tuple(param)

    else:
        # create relevant list of parameters from dict
        str_keys = [key for key, val in params.items() if comp_str in key]
        # print(str_keys, 'str_keys')
        param = []
        for j in range(len(str_keys)):
            if str.startswith(comp_str, 'bg'):
                param.append(params[comp_str + '_f' + str(j)])
            else:
                param.append(params[comp_str + str(j)])

    fout = np.ones(azimu.shape)
    # print(azimu, fout, 'azimu', type(param))
    if azimu.size == 1:  # this line is required to catch error when out is single number.
        try:
            fout = param[0]
        except IndexError:
            fout = param
    else:
        fout[:] = param[0]
    # essentially d_0, h_0 or w_0

    if not isinstance(param, np.float64) and np.size(param) > 1:
        for i in range(1, int((len(param) - 1) / 2) + 1):
            # len(param)-1 should never be odd because of initial a_0 parameter
            fout = fout + param[(2 * i) - 1] * np.sin(np.deg2rad(azimu) * i) + param[2 * i] * np.cos(
                np.deg2rad(azimu) * i)  # single col array
    # else:
    #  azimu = np.array(azimu)
    # fout = np.ones(azimu.shape)
    # fout[:] = out
    # out = fout*param

    return fout


def Fourier_fit(ydata, azimu, param, terms=None, errs=None, param_str='peak_0', symm=1, fit_method='leastsq'):
    """Fit the Fourier expansion to number of required terms
    :param ydata: Component data array to fit float
    :param azimu: data array float
    :param param: lmfit Parameter class dict
    :param terms: number of terms to fit int
    :param errs: Component data array errors of size ydata
    :param param_str: str start of parameter name
    :param symm: symmetry in azimuth of the fourier to be fit
    :param fit_method: lmfit method default 'leastsq'
    :return: lmfit Model result
    """

    # get NaN values.
    idx = np.isfinite(azimu) & np.isfinite(ydata) 
    if type(terms) == list:
        terms = terms[0]
    if param:
        param = param
    else:
        param = Parameters()
        for t in range(2 * terms + 1):
            params.add(param_str + '_' + str(t), 1.)  # need limits

    if errs is None or errs.all is None:
        errs = np.ones(ydata.shape)

    # param.pretty_print()
    fmodel = Model(Fourier_expand, independent_vars=['azimu'])
    # print('parameter names: {}'.format(fmodel.param_names))
    # print('independent variables: {}'.format(fmodel.independent_vars))
    # Attempt to mitigate failure of fit with weights containing 'None' values
    # Replace with nan and alter dtype from object to float64
    new_errs = errs[idx]
    new_errs[new_errs == None] = 1000 * new_errs[new_errs != None].max()
    new_errs = new_errs.astype('float64')
    out = fmodel.fit(ydata[idx], param, azimu=azimu[idx]*symm, method=fit_method, sigma=new_errs, comp_str=param_str,
                     nan_policy='propagate')

    return out


# fourier expansion function
def Fourier_backgrnd(azimutheta, param):
    """
    Calculate the Fourier expansion of the background terms
    :param azimutheta: list of arrays for azimuth and theta float
    :param param: list of input parameters
    :return: Fourier expansion result float arr
    """
    azimu, twotheta = azimutheta
    twothetaprime = twotheta - twotheta.min()
    backg = param
    bg_all = np.zeros(twotheta.shape)
    # print(backg, bg_all)

    # Not sure if this is correct, thought that if orders then array would be eg [5,0] and would want a fourier
    # expansion with 5 parms for offset and then no fourier expansion for slope i.e. 5 parms then 0.
    # But below would interpret this as effectively [0,0] instead.

    for i in range(len(backg)):
        try:
            nterms = len(backg[i])
        except TypeError:
            nterms = backg[i].size
        # print((nterms - 1) / 2)
        out = backg[i][0]
        for j in range(1, int((nterms - 1) / 2) + 1):
            out = out + backg[i][(2 * j) - 1] * np.sin(np.deg2rad(azimu) * j) + backg[i][2 * j] * np.cos(
                np.deg2rad(azimu) * j)
        bg_all = bg_all + (out * (twothetaprime ** float(i)))

    return bg_all
